"""Block Neural Autoregressive bijection implementation."""
from typing import Callable

import jax
import jax.numpy as jnp
from jax import random
from jax.random import KeyArray

from flowjax.bijections.bijection import Bijection
from flowjax.nn.block_autoregressive import (
    BlockAutoregressiveLinear,
    _block_tanh_activation,
)


class BlockAutoregressiveNetwork(Bijection):
    """Block Autoregressive Network (https://arxiv.org/abs/1904.04676)."""

    depth: int
    layers: list
    block_dim: int
    activation: Callable

    def __init__(
        self,
        key: KeyArray,
        dim: int,
        cond_dim: int | None,
        depth: int,
        block_dim: int,
        activation: Callable | None = None,
    ):
        """
        Args:
            key (KeyArray): Jax PRNGKey
            dim (int): Dimension of the distribution.
            cond_dim (tuple[int, ...] | None): Dimension of conditioning variables.
            depth (int): Number of hidden layers in the network.
            block_dim (int): Block dimension (hidden layer size is `dim*block_dim`).
            activation (Callable): Activation function. Defaults to block_tanh.
        """
        activation = _block_tanh_activation(dim) if activation is None else activation
        layers = []
        if depth == 0:
            layers.append(BlockAutoregressiveLinear(key, dim, (1, 1), cond_dim))
        else:
            keys = random.split(key, depth + 1)

            block_shapes = [
                (block_dim, 1),
                *[(block_dim, block_dim)] * (depth - 1),
                (1, block_dim),
            ]
            cond_dims = [cond_dim] + [None] * depth

            for layer_key, block_shape, cond_d in zip(
                keys, block_shapes, cond_dims, strict=True
            ):
                layers.extend(
                    [
                        BlockAutoregressiveLinear(layer_key, dim, block_shape, cond_d),
                        activation,
                    ]
                )
            layers = layers[:-1]  # remove last activation

        self.depth = depth
        self.layers = layers
        self.block_dim = block_dim
        self.activation = activation
        self.shape = (dim,)
        self.cond_shape = (cond_dim,) if cond_dim is not None else None

    def transform(self, x, condition=None):
        x, condition = self._argcheck_and_cast(x, condition)
        x = self.layers[0](x, condition)[0]
        for layer in self.layers[1:]:
            x = layer(x)[0]
        return x

    def transform_and_log_det(self, x, condition=None):
        x, condition = self._argcheck_and_cast(x, condition)
        x, log_jacobian_3d_0 = self.layers[0](x, condition)
        log_jacobian_3ds = [log_jacobian_3d_0]
        for layer in self.layers[1:]:
            x, log_det_3d = layer(x)
            log_jacobian_3ds.append(log_det_3d)

        log_det = log_jacobian_3ds[-1]
        for log_jacobian in reversed(log_jacobian_3ds[:-1]):
            log_det = logmatmulexp(log_det, log_jacobian)
        return x, log_det.sum()

    def inverse(self, *args, **kwargs):
        raise NotImplementedError(
            "This transform would require numerical methods for inversion."
        )

    def inverse_and_log_det(self, *args, **kwargs):
        raise NotImplementedError(
            "This transform would require numerical methods for inversion."
        )


def logmatmulexp(x, y):
    """
    Numerically stable version of ``(x.log() @ y.log()).exp()``.
    From numpyro https://github.com/pyro-ppl/numpyro/blob/
    f2ff89a3a7147617e185eb51148eb15d56d44661/numpyro/distributions/util.py#L387
    """
    x_shift = jax.lax.stop_gradient(jnp.amax(x, -1, keepdims=True))
    y_shift = jax.lax.stop_gradient(jnp.amax(y, -2, keepdims=True))
    xy = jnp.log(jnp.matmul(jnp.exp(x - x_shift), jnp.exp(y - y_shift)))
    return xy + x_shift + y_shift



# Configuration file for the Sphinx documentation builder.
import os
import sys

import jax  # noqa Avoid circular import

# Avoid unused module linting

sys.path.insert(0, os.path.abspath(".."))

# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "FlowJax"
copyright = "2022, Daniel Ward"
author = "Daniel Ward"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.viewcode",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.doctest",
    "nbsphinx",
    "sphinx_copybutton",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

add_module_names = False
napoleon_include_init_with_doc = True

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]

html_css_files = [
    "style.css",
]


html_theme_options = {
    "navigation_depth": 2,
}

pygments_style = "xcode"
autodoc_typehints = "none"

copybutton_prompt_text = r">>> |\.\.\. "
copybutton_prompt_is_regexp = True



"""Planar transform, i.e. a layer in a planar flow introduced in
https://arxiv.org/pdf/1505.05770.pdf.
"""

import equinox as eqx
import jax.numpy as jnp
import jax.random as jr
from jax import Array
from jax.nn import softplus
from jax.numpy.linalg import norm

from flowjax.bijections import Bijection


class Planar(Bijection):
    r"""Planar bijection as used by https://arxiv.org/pdf/1505.05770.pdf. Uses the
    transformation :math:`y + u \cdot \text{tanh}(w \cdot x + b)`, where
    :math:`u \in \mathbb{R}^D, \ w \in \mathbb{R}^D` and :math:`b \in \mathbb{R}`. In
    the unconditional case, :math:`w`, :math:`u`  and :math:`b` are learned directly.
    In the conditional case they are parameterised by an MLP.
    """

    conditioner: eqx.Module | None
    params: Array | None

    def __init__(
        self,
        key: jr.KeyArray,
        dim: int,
        cond_dim: int | None = None,
        **mlp_kwargs,
    ):
        """
        Args:
            key (jr.KeyArray): Jax random seed.
            dim (int): Dimension of the bijection.
            cond_dim (int | None, optional): Dimension of extra conditioning variables.
                Defaults to None.
            **mlp_kwargs: Key word arguments passed to the MLP conditioner. Ignored
                when cond_dim is None.
        """
        self.shape = (dim,)

        if cond_dim is None:
            self.params = 0.01 * jr.normal(key, (2 * dim + 1,))
            self.conditioner = None
            self.cond_shape = None
        else:
            self.params = None
            self.conditioner = eqx.nn.MLP(dim, 2 * dim + 1, **mlp_kwargs, key=key)
            self.cond_shape = (cond_dim,)

    def transform(self, x, condition=None):
        x, condition = self._argcheck_and_cast(x, condition)
        return self.get_planar(condition).transform(x)

    def transform_and_log_det(self, x, condition=None):
        x, condition = self._argcheck_and_cast(x, condition)
        return self.get_planar(condition).transform_and_log_det(x)

    def inverse(self, y, condition=None):
        return self.get_planar(condition).inverse(y)

    def inverse_and_log_det(self, y, condition=None):
        return self.get_planar(condition).inverse_and_log_det(y)

    def get_planar(self, condition=None):
        "Get the planar bijection with the conditioning applied if conditional."
        if self.cond_shape is not None:
            params = self.conditioner(condition)
        else:
            params = self.params
        dim = self.shape[0]
        w, u, bias = params[:dim], params[dim : 2 * dim], params[-1]
        return _UnconditionalPlanar(w, u, bias)


class _UnconditionalPlanar(Bijection):
    """Unconditional planar bijection, used in Planar."""

    weight: Array
    _act_scale: Array
    bias: Array

    def __init__(self, weight, act_scale, bias):
        """Construct an unconditional planar bijection. Note act_scale (u in the paper)
        is unconstrained and the constraint to ensure invertiblitiy is applied in the
        ``get_act_scale``."""
        self.weight = weight
        self._act_scale = act_scale
        self.bias = bias
        self.shape = weight.shape
        self.cond_shape = None

    def transform(self, x, condition=None):
        return x + self.get_act_scale() * jnp.tanh(self.weight @ x + self.bias)

    def transform_and_log_det(self, x, condition=None):
        u = self.get_act_scale()
        act = jnp.tanh(x @ self.weight + self.bias)
        y = x + u * act
        psi = (1 - act**2) * self.weight
        log_det = jnp.log(jnp.abs(1 + u @ psi))
        return y, log_det

    def get_act_scale(self):
        """Apply constraint to u to ensure invertibility. See appendix A1 in
        https://arxiv.org/pdf/1505.05770.pdf."""
        wtu = self._act_scale @ self.weight
        m_wtu = -1 + jnp.log(1 + softplus(wtu))
        u = self._act_scale + (m_wtu - wtu) * self.weight / norm(self.weight) ** 2
        return u

    def inverse(self, y, condition=None):
        raise NotImplementedError(
            "The inverse planar transformation is not implemented."
        )

    def inverse_and_log_det(self, y, condition=None):
        raise NotImplementedError(
            "The inverse planar transformation is not implemented."
        )



"""Autoregressive linear layers and multilayer perceptron."""
from typing import Callable

import jax.nn as jnn
import jax.numpy as jnp
from equinox import Module
from equinox.nn import Linear
from jax import Array, random
from jax.random import KeyArray
from jax.typing import ArrayLike

from flowjax.masks import rank_based_mask


def _identity(x):
    return x


class MaskedLinear(Module):
    """Masked linear neural network layer."""

    linear: Linear
    mask: Array

    def __init__(self, mask: ArrayLike, use_bias: bool = True, *, key: KeyArray):
        """

        Args:
            mask (ArrayLike): Mask with shape (out_features, in_features).
            key (KeyArray): Jax PRNGKey
            use_bias (bool): Whether to include bias terms. Defaults to True.
        """
        mask = jnp.asarray(mask)
        self.linear = Linear(mask.shape[1], mask.shape[0], use_bias, key=key)
        self.mask = mask

    def __call__(self, x: ArrayLike):
        """Run the masked linear layer

        Args:
            x (ArrayLike): Array with shape ``(mask.shape[1], )``
        """
        x = jnp.asarray(x)
        x = self.linear.weight * self.mask @ x
        if self.linear.bias is not None:
            x = x + self.linear.bias
        return x


class AutoregressiveMLP(Module):
    """An autoregressive multilayer perceptron, similar to ``equinox.nn.composed.MLP``.
    Connections will only exist where in_ranks < out_ranks.
    """

    in_size: int
    out_size: int
    width_size: int
    depth: int
    in_ranks: Array
    out_ranks: Array
    hidden_ranks: Array
    layers: list[MaskedLinear]
    activation: Callable
    final_activation: Callable

    def __init__(
        self,
        in_ranks: ArrayLike,
        hidden_ranks: ArrayLike,
        out_ranks: ArrayLike,
        depth: int,
        activation: Callable = jnn.relu,
        final_activation: Callable = _identity,
        *,
        key,
    ) -> None:
        """
        Args:
            in_ranks (ArrayLike): Ranks of the inputs.
            hidden_ranks (ArrayLike): Ranks of the hidden layer(s).
            out_ranks (ArrayLike): Ranks of the outputs.
            depth (int): Number of hidden layers.
            activation (Callable): Activation function. Defaults to jnn.relu.
            final_activation (Callable): Final activation function. Defaults to
                _identity.
            key (KeyArray): Jax PRNGKey.
        """
        in_ranks, hidden_ranks, out_ranks = [
            jnp.asarray(a) for a in [in_ranks, hidden_ranks, out_ranks]
        ]
        masks = []
        if depth == 0:
            masks.append(rank_based_mask(in_ranks, out_ranks, eq=False))
        else:
            masks.append(rank_based_mask(in_ranks, hidden_ranks, eq=True))
            for _ in range(depth - 1):
                masks.append(rank_based_mask(hidden_ranks, hidden_ranks, eq=True))
            masks.append(rank_based_mask(hidden_ranks, out_ranks, eq=False))

        keys = random.split(key, len(masks))
        layers = [
            MaskedLinear(mask, key=key) for mask, key in zip(masks, keys, strict=True)
        ]

        self.layers = layers
        self.in_size = len(in_ranks)
        self.out_size = len(out_ranks)
        self.width_size = len(hidden_ranks)
        self.depth = depth
        self.in_ranks = in_ranks
        self.hidden_ranks = hidden_ranks
        self.out_ranks = out_ranks
        self.activation = activation
        self.final_activation = final_activation

    def __call__(self, x: Array):
        """Forward pass.
        Args:
            x: A JAX array with shape (in_size,).
        """
        for layer in self.layers[:-1]:
            x = layer(x)
            x = self.activation(x)
        x = self.layers[-1](x)
        x = self.final_activation(x)
        return x



"""Implemenetation of Coupling flow layer with arbitrary transformer.
See https://arxiv.org/abs/1605.08803 for more information.
"""
from typing import Callable

import equinox as eqx
import jax.nn as jnn
import jax.numpy as jnp
from jax.random import KeyArray

from flowjax.bijections.bijection import Bijection
from flowjax.bijections.jax_transforms import Batch
from flowjax.utils import Array, get_ravelled_bijection_constructor


class Coupling(Bijection):
    """Coupling layer implementation (https://arxiv.org/abs/1605.08803)."""

    untransformed_dim: int
    dim: int
    transformer_constructor: Callable
    conditioner: eqx.nn.MLP

    def __init__(
        self,
        key: KeyArray,
        transformer: Bijection,
        untransformed_dim: int,
        dim: int,
        cond_dim: int | None,
        nn_width: int,
        nn_depth: int,
        nn_activation: Callable = jnn.relu,
    ):
        """
        Args:
            key (KeyArray): Jax PRNGKey
            transformer (Bijection): Unconditional bijection with shape () to be
                parameterised by the conditioner neural netork.
            untransformed_dim (int): Number of untransformed conditioning variables (
                e.g. dim // 2).
            dim (int): Total dimension.
            cond_dim (int | None): Dimension of additional conditioning variables.
            nn_width (int): Neural network hidden layer width.
            nn_depth (int): Neural network hidden layer size.
            nn_activation (Callable): Neural network activation function.
                Defaults to jnn.relu.
        """
        if transformer.shape != () or transformer.cond_shape is not None:
            raise ValueError(
                "Only unconditional transformers with shape () are supported."
            )

        constructor, transformer_init_params = get_ravelled_bijection_constructor(
            transformer
        )

        self.transformer_constructor = constructor
        self.untransformed_dim = untransformed_dim
        self.dim = dim
        self.shape = (dim,)
        self.cond_shape = (cond_dim,) if cond_dim is not None else None

        conditioner_output_size = transformer_init_params.size * (
            dim - untransformed_dim
        )

        conditioner = eqx.nn.MLP(
            in_size=untransformed_dim
            if cond_dim is None
            else untransformed_dim + cond_dim,
            out_size=conditioner_output_size,
            width_size=nn_width,
            depth=nn_depth,
            activation=nn_activation,
            key=key,
        )  # type: eqx.nn.MLP

        # Initialise last bias terms to match the provided transformer parameters
        self.conditioner = eqx.tree_at(
            where=lambda mlp: mlp.layers[-1].bias,  # type: ignore
            pytree=conditioner,
            replace=jnp.tile(transformer_init_params, dim - untransformed_dim),
        )

    def transform(self, x, condition=None):
        x_cond, x_trans = x[: self.untransformed_dim], x[self.untransformed_dim :]
        nn_input = x_cond if condition is None else jnp.hstack((x_cond, condition))
        transformer_params = self.conditioner(nn_input)
        transformer = self._flat_params_to_transformer(transformer_params)
        y_trans = transformer.transform(x_trans)
        y = jnp.hstack((x_cond, y_trans))
        return y

    def transform_and_log_det(self, x, condition=None):
        x_cond, x_trans = x[: self.untransformed_dim], x[self.untransformed_dim :]
        nn_input = x_cond if condition is None else jnp.hstack((x_cond, condition))
        transformer_params = self.conditioner(nn_input)
        transformer = self._flat_params_to_transformer(transformer_params)
        y_trans, log_det = transformer.transform_and_log_det(x_trans)
        y = jnp.hstack((x_cond, y_trans))
        return y, log_det

    def inverse(self, y, condition=None):
        x_cond, y_trans = y[: self.untransformed_dim], y[self.untransformed_dim :]
        nn_input = x_cond if condition is None else jnp.concatenate((x_cond, condition))
        transformer_params = self.conditioner(nn_input)
        transformer = self._flat_params_to_transformer(transformer_params)
        x_trans = transformer.inverse(y_trans)
        x = jnp.hstack((x_cond, x_trans))
        return x

    def inverse_and_log_det(self, y, condition=None):
        x_cond, y_trans = y[: self.untransformed_dim], y[self.untransformed_dim :]
        nn_input = x_cond if condition is None else jnp.concatenate((x_cond, condition))
        transformer_params = self.conditioner(nn_input)
        transformer = self._flat_params_to_transformer(transformer_params)
        x_trans, log_det = transformer.inverse_and_log_det(y_trans)
        x = jnp.hstack((x_cond, x_trans))
        return x, log_det

    def _flat_params_to_transformer(self, params: Array):
        """Reshape to dim X params_per_dim, then vmap."""
        dim = self.dim - self.untransformed_dim
        transformer_params = jnp.reshape(params, (dim, -1))
        transformer = eqx.filter_vmap(self.transformer_constructor)(transformer_params)
        return Batch(transformer, (dim,), vectorize_bijection=True)



import jax.numpy as jnp
import jax.random as jr
import numpy as np
import pytest

from flowjax.distributions import (
    Cauchy,
    Distribution,
    Gumbel,
    Normal,
    StandardNormal,
    StudentT,
    Uniform,
    _StandardCauchy,
    _StandardGumbel,
    _StandardStudentT,
    _StandardUniform,
)

# This sets up a number of constructors shape -> instance for testing
# the generic API of Distribution classes.
# Note we do not test the private "Standard" distributions, assuming they are
# sufficiently tested by their loc, scale public counterparts.

_test_distributions = {
    # flowjax.distributions
    "StandardNormal": lambda shape: StandardNormal(shape),
    "Normal": lambda shape: Normal(jnp.zeros(shape)),
    "_StandardUniform": lambda shape: _StandardUniform(shape),
    "Uniform": lambda shape: Uniform(jnp.zeros(shape), 1),
    "_StandardGumbel": _StandardGumbel,
    "Gumbel": lambda shape: Gumbel(jnp.zeros(shape)),
    "_StandardCauchy": _StandardCauchy,
    "Cauchy": lambda shape: Cauchy(jnp.zeros(shape)),
    "_StandardStudentT": lambda shape: _StandardStudentT(jnp.ones(shape)),
    "StudentT": lambda shape: StudentT(jnp.ones(shape)),
}


_test_distributions = [pytest.param(v, id=k) for k, v in _test_distributions.items()]
_test_shapes = [(), (2,), (2, 3)]


@pytest.mark.parametrize("distribution", _test_distributions)
@pytest.mark.parametrize("shape", _test_shapes)
def test_sample(distribution, shape):
    d = distribution(shape=shape)
    sample = d.sample(jr.PRNGKey(0))
    assert sample.shape == shape

    sample_shape = (1, 2)
    sample = d.sample(jr.PRNGKey(0), sample_shape)
    assert sample.shape == sample_shape + shape


@pytest.mark.parametrize("distribution", _test_distributions)
@pytest.mark.parametrize("shape", _test_shapes)
def test_log_prob(distribution, shape):
    d = distribution(shape=shape)
    x = d.sample(jr.PRNGKey(0))

    assert d.log_prob(x).shape == ()

    sample_shape = (1, 2)
    x = d.sample(jr.PRNGKey(0), sample_shape)
    assert d.log_prob(x).shape == sample_shape

    # test arraylike input
    assert jnp.all(d.log_prob(np.array(x)) == d.log_prob(x))

    if d.shape == ():
        assert d.log_prob(jnp.array(0)) == d.log_prob(0)


@pytest.mark.parametrize("distribution", _test_distributions)
def test_log_prob_shape_mismatch(distribution):
    d = distribution(shape=(3,))

    with pytest.raises(ValueError):
        d.log_prob(jnp.ones((3, 2)))

    d = distribution(shape=(3, 2))
    with pytest.raises(ValueError):
        d.log_prob(jnp.ones((2,)))


def test_uniform_params():
    dist = Uniform(
        jnp.array([1.0, 2.0]),
        jnp.array([3.0, 4.0]),
    )

    assert dist.minval == pytest.approx(jnp.array([1.0, 2.0]))
    assert dist.maxval == pytest.approx(jnp.array([3.0, 4.0]))


# Since the broadcasting behaviour is shared by all distributions
# we test it for a single unconditional and conditional distribution only.

dist_shape, sample_shape, condition_shape = [[(), (2,), (3, 4)] for _ in range(3)]


class _TestDist(Distribution):
    "Toy distribution object, for testing of distribution broadcasting."

    def __init__(self, shape, cond_shape=None) -> None:
        self.shape = shape
        self.cond_shape = cond_shape

    def _log_prob(self, x, condition=None):
        return jnp.zeros(())

    def _sample(self, key, condition=None):
        return jnp.zeros(self.shape)


@pytest.mark.parametrize("dist_shape", dist_shape)
@pytest.mark.parametrize("sample_shape", sample_shape)
def test_broadcasting_unconditional(dist_shape, sample_shape):
    d = _TestDist(dist_shape)
    samples = d.sample(jr.PRNGKey(0), sample_shape)
    assert samples.shape == sample_shape + dist_shape

    log_probs = d.log_prob(samples)
    assert log_probs.shape == sample_shape

    with pytest.raises(ValueError):
        d.sample(jr.PRNGKey(0), sample_shape, condition=jnp.ones(3))

    with pytest.raises(ValueError):
        d.log_prob(samples, condition=jnp.ones(3))


@pytest.mark.parametrize("dist_shape", dist_shape)
@pytest.mark.parametrize("sample_shape", sample_shape)
@pytest.mark.parametrize("condition_shape", condition_shape)
@pytest.mark.parametrize(
    "leading_cond_shape", [(), (3, 4)]
)  # Additional leading dimensions in condition
def test_broadcasting_conditional(
    dist_shape, sample_shape, condition_shape, leading_cond_shape
):
    key = jr.PRNGKey(0)
    d = _TestDist(dist_shape, condition_shape)
    condition = jnp.zeros(leading_cond_shape + condition_shape)
    samples = d.sample(key, sample_shape, condition)
    assert samples.shape == sample_shape + leading_cond_shape + dist_shape

    log_probs = d.log_prob(samples, condition)
    assert log_probs.shape == sample_shape + leading_cond_shape

    samples, log_probs = d.sample_and_log_prob(key, sample_shape, condition)
    assert samples.shape == sample_shape + leading_cond_shape + dist_shape
    assert log_probs.shape == sample_shape + leading_cond_shape


test_cases = [
    StandardNormal(
        (2, 2)
    ),  # Won't have custom sample_and_log_prob implementation as not Transformed
    Normal(jnp.ones((2, 2))),  # Will have custom implementation as is Transformed
]


@pytest.mark.parametrize("dist", test_cases)
def test_sample_and_log_prob(dist):
    # We test broadcasting behaviour seperately above.
    # Just check consistency to seperately using methods
    key = jr.PRNGKey(0)
    x_naive = dist._sample(key)
    lp_naive = dist._log_prob(x_naive)
    x, lp = dist._sample_and_log_prob(key)
    assert x == pytest.approx(x_naive)
    assert lp == pytest.approx(lp_naive)



import jax.random as jr
import pytest

from flowjax.bijections import Affine, RationalQuadraticSpline
from flowjax.distributions import StandardNormal
from flowjax.flows import (
    BlockNeuralAutoregressiveFlow,
    CouplingFlow,
    MaskedAutoregressiveFlow,
    TriangularSplineFlow,
)

DIM = 3
KEY = jr.PRNGKey(0)
KWARGS = {
    "key": jr.PRNGKey(0),
    "base_dist": StandardNormal((DIM,)),
    "flow_layers": 2,
}

testcases = {
    "BNAF": BlockNeuralAutoregressiveFlow(**KWARGS),
    "TriangularSplineFlow": TriangularSplineFlow(**KWARGS),
    "Affine_Coupling": CouplingFlow(transformer=Affine(), **KWARGS),
    "Spline_Coupling": CouplingFlow(
        transformer=RationalQuadraticSpline(3, 2), **KWARGS
    ),
    "Affine_MaskedAutoregessive": MaskedAutoregressiveFlow(
        transformer=Affine(), **KWARGS
    ),
}


@pytest.mark.parametrize("flow", testcases.values(), ids=testcases.keys())
def test_unconditional_flow_sample(flow):
    try:
        assert flow._sample(KEY).shape == flow.shape
    except NotImplementedError:
        pass


@pytest.mark.parametrize("flow", testcases.values(), ids=testcases.keys())
def test_unconditional_flow_log_prob(flow):
    x = jr.normal(KEY, flow.shape)
    assert flow._log_prob(x).shape == ()


conditional_testcases = {
    "BNAF": BlockNeuralAutoregressiveFlow(**KWARGS, cond_dim=2),
    "TriangularSplineFlow": TriangularSplineFlow(**KWARGS, cond_dim=2),
    "Affine_Coupling": CouplingFlow(transformer=Affine(), **KWARGS, cond_dim=2),
    "Spline_Coupling": CouplingFlow(
        transformer=RationalQuadraticSpline(3, 2), **KWARGS, cond_dim=2
    ),
    "Affine_MaskedAutoregessive": MaskedAutoregressiveFlow(
        transformer=Affine(), **KWARGS, cond_dim=2
    ),
}


@pytest.mark.parametrize(
    "flow", conditional_testcases.values(), ids=conditional_testcases.keys()
)
def test_conditional_flow_sample(flow):
    cond = jr.normal(KEY, flow.cond_shape)
    try:
        assert flow._sample(KEY, cond).shape == flow.shape
    except NotImplementedError:
        pass


@pytest.mark.parametrize(
    "flow", conditional_testcases.values(), ids=conditional_testcases.keys()
)
def test_conditional_flow_log_prob(flow):
    x = jr.normal(KEY, flow.shape)
    cond = jr.normal(KEY, flow.cond_shape)
    assert flow._log_prob(x, cond).shape == ()



"""Module contains bijections formed by "stacking/concatenating" other bijections."""

from typing import Sequence

import jax.numpy as jnp
from jax import Array

from flowjax.bijections.bijection import Bijection
from flowjax.utils import check_shapes_match, merge_cond_shapes


class Concatenate(Bijection):
    """Concatenate bijections along an already existing axis. Analagous to
    ``jnp.concatenate``. See also :class:`Stack`.
    """

    split_idxs: Array
    bijections: Sequence[Bijection]
    axis: int

    def __init__(self, bijections: Sequence[Bijection], axis: int = 0):
        """
        Args:
            bijections (Sequence[Bijection]): Bijections, to stack into a single
                bijection.
            axis (int): Axis along which to stack. Defaults to 0.
        """
        self.bijections = bijections
        self.axis = axis

        shapes = [b.shape for b in bijections]
        self._argcheck_shapes(shapes)
        axis = range(len(shapes[0]))[axis]  # Avoids issues when axis==-1
        self.shape = (
            shapes[0][:axis] + (sum(s[axis] for s in shapes),) + shapes[0][axis + 1 :]
        )
        self.split_idxs = jnp.array([s[axis] for s in shapes[:-1]])
        self.cond_shape = merge_cond_shapes([b.cond_shape for b in bijections])

    def transform(self, x, condition=None):
        x, condition = self._argcheck_and_cast(x, condition)
        x_parts = jnp.array_split(x, self.split_idxs, axis=self.axis)
        y_parts = [
            b.transform(x_part, condition)
            for b, x_part in zip(self.bijections, x_parts, strict=True)
        ]
        return jnp.concatenate(y_parts, axis=self.axis)

    def transform_and_log_det(self, x, condition=None):
        x, condition = self._argcheck_and_cast(x, condition)
        x_parts = jnp.array_split(x, self.split_idxs, axis=self.axis)

        ys_log_dets = [
            b.transform_and_log_det(x, condition)
            for b, x in zip(self.bijections, x_parts, strict=True)
        ]

        y_parts, log_dets = zip(*ys_log_dets, strict=True)
        return jnp.concatenate(y_parts, self.axis), sum(log_dets)

    def inverse(self, y, condition=None):
        y, condition = self._argcheck_and_cast(y, condition)
        y_parts = jnp.array_split(y, self.split_idxs, axis=self.axis)
        x_parts = [
            b.inverse(y_part, condition)
            for b, y_part in zip(self.bijections, y_parts, strict=True)
        ]
        return jnp.concatenate(x_parts, axis=self.axis)

    def inverse_and_log_det(self, y, condition=None):
        y, condition = self._argcheck_and_cast(y, condition)
        y_parts = jnp.array_split(y, self.split_idxs, axis=self.axis)

        xs_log_dets = [
            b.inverse_and_log_det(y, condition)
            for b, y in zip(self.bijections, y_parts, strict=True)
        ]

        x_parts, log_dets = zip(*xs_log_dets, strict=True)
        return jnp.concatenate(x_parts, self.axis), sum(log_dets)

    def _argcheck_shapes(self, shapes: list[tuple[int, ...]]):
        axis = range(len(shapes[0]))[self.axis]

        for i, shp in enumerate(shapes):
            if shp[:axis] + shp[axis + 1 :] != shapes[0][:axis] + shapes[0][axis + 1 :]:
                raise ValueError(
                    f"Expected bijection shapes to match except along axis {axis}, but "
                    f"index 0 had shape {shapes[0]}, and index {i} had shape {shp}."
                )


class Stack(Bijection):
    """
    Stack bijections along a new axis (analagous to ``jnp.stack``).
    See also :class:`Concatenate`.
    """

    bijections: Sequence[Bijection]
    axis: int

    def __init__(self, bijections: list[Bijection], axis: int = 0):
        """
        Args:
            bijections (list[Bijection]): Bijections.
            axis (int): Axis along which to stack. Defaults to 0.
        """
        self.axis = axis
        self.bijections = bijections

        shapes = [b.shape for b in bijections]
        check_shapes_match(shapes)

        self.shape = shapes[0][:axis] + (len(bijections),) + shapes[0][axis:]
        self.cond_shape = merge_cond_shapes([b.cond_shape for b in bijections])

    def transform(self, x, condition=None):
        x, condition = self._argcheck_and_cast(x, condition)
        x_parts = self._split_and_squeeze(x)
        y_parts = [
            b.transform(x, condition)
            for (b, x) in zip(self.bijections, x_parts, strict=True)
        ]
        return jnp.stack(y_parts, self.axis)

    def transform_and_log_det(self, x, condition=None):
        x, condition = self._argcheck_and_cast(x, condition)
        x_parts = self._split_and_squeeze(x)
        ys_log_det = [
            b.transform_and_log_det(x, condition)
            for b, x in zip(self.bijections, x_parts, strict=True)
        ]

        y_parts, log_dets = zip(*ys_log_det, strict=True)
        return jnp.stack(y_parts, self.axis), sum(log_dets)

    def inverse(self, y, condition=None):
        y, condition = self._argcheck_and_cast(y, condition)
        y_parts = self._split_and_squeeze(y)
        x_parts = [
            b.inverse(y, condition)
            for (b, y) in zip(self.bijections, y_parts, strict=True)
        ]
        return jnp.stack(x_parts, self.axis)

    def inverse_and_log_det(self, y, condition=None):
        y, condition = self._argcheck_and_cast(y, condition)
        y_parts = self._split_and_squeeze(y)
        xs_log_det = [
            b.inverse_and_log_det(y, condition)
            for b, y in zip(self.bijections, y_parts, strict=True)
        ]
        x_parts, log_dets = zip(*xs_log_det, strict=True)
        return jnp.stack(x_parts, self.axis), sum(log_dets)

    def _split_and_squeeze(self, array: Array):
        arrays = jnp.split(array, len(self.bijections), axis=self.axis)
        return [a.squeeze(axis=self.axis) for a in arrays]



"Tests for bijection.chain module"
from functools import partial

import equinox as eqx
import jax.numpy as jnp
import pytest
from jax import random

from flowjax.bijections import Affine, Chain, Coupling, Flip, Permute, Scan


def test_chain_dunders():
    bijection = Chain([Flip((2,)), Permute(jnp.array([0, 1]))])
    assert len(bijection) == 2
    assert isinstance(bijection[0], Flip)
    assert isinstance(bijection[1], Permute)
    assert isinstance(bijection[:], Chain)


DIM = 4
COND_DIM = 5
NUM_LAYERS = 3
keys = random.split(random.PRNGKey(0), NUM_LAYERS)

make_coupling_layer = partial(
    Coupling,
    transformer=Affine(),
    untransformed_dim=DIM // 2,
    dim=DIM,
    cond_dim=COND_DIM,
    nn_width=10,
    nn_depth=3,
)


params = jnp.ones((NUM_LAYERS, DIM))
affine_chain = Chain([Affine(p) for p in params])
affine_scan = Scan(eqx.filter_vmap(Affine)(params))

coupling_chain = Chain([make_coupling_layer(k) for k in keys])
coupling_scan = Scan(eqx.filter_vmap(make_coupling_layer)(keys))

test_cases = {
    "Affine": (affine_scan, affine_chain),
    "Coupling": (coupling_scan, coupling_chain),
}


@pytest.mark.parametrize("scan,chain", test_cases.values(), ids=test_cases.keys())
def test_scan(scan, chain):
    "Check Chain and Scan give consistent results."
    x = jnp.ones(DIM)
    condition = jnp.ones(chain.cond_shape) if chain.cond_shape is not None else None
    expected = pytest.approx(chain.transform(x, condition))
    assert expected == scan.transform(x, condition)

    expected = pytest.approx(chain.inverse(x, condition))
    realised = scan.inverse(x, condition)
    assert expected == realised

    expected = chain.transform_and_log_det(x, condition)
    realised = scan.transform_and_log_det(x, condition)
    assert jnp.all(
        jnp.array(
            [pytest.approx(a) == b for (a, b) in zip(expected, realised, strict=True)]
        )
    )



"""Distributions - including the base class Distribution, common distributions
and a Transformed distribution class.
"""
from abc import abstractmethod
from math import prod

import equinox as eqx
import jax.numpy as jnp
import jax.random as jr
from jax import Array
from jax.experimental import checkify
from jax.lax import stop_gradient
from jax.scipy import stats as jstats
from jax.typing import ArrayLike

from flowjax.bijections import Affine, Bijection
from flowjax.utils import _get_ufunc_signature, arraylike_to_array, merge_cond_shapes


class Distribution(eqx.Module):
    """Distribution base class. Distributions all have an attribute ``shape``,
    denoting the shape of a single sample from the distribution. This corresponds to the
    ``batch_shape + event_shape`` in torch/numpyro distributions. Similarly, the
    ``cond_shape`` attribute denotes the shape of the conditioning variable.
    This attribute is None for unconditional distributions. For example

    .. doctest::

        >>> import jax.numpy as jnp
        >>> from flowjax.distributions import Normal
        >>> dist = Normal(jnp.zeros(2))
        >>> dist.shape
        (2,)
        >>> dist.cond_shape is None
        True

    Distributions are registered as jax PyTrees (as they are equinox modules), and as
    such they are compatible with normal jax operations.

    **Implementing a distribution**

        (1) Inherit from ``Distribution``.
        (2) Define attributes ``shape`` and ``cond_shape``. ``cond_shape`` should be
            ``None`` for unconditional distributions.
        (3) Define the ``_sample`` method, which samples a point with a shape of
            ``shape``, (given a conditioning variable with shape ``cond_shape`` for
            conditional distributions).
        (4) Define the ``_log_prob`` method, which evaluates the log probability,
            given an input of shape ``shape`` (and a conditioning variable with shape
            ``cond_shape`` for conditional distributions).

        The base class will handle defining more convenient log_prob and sample methods
        that support broadcasting and perform argument checks.

    """

    shape: tuple[int, ...]
    cond_shape: tuple[int, ...] | None

    @abstractmethod
    def _log_prob(self, x: Array, condition: Array | None = None) -> Array:
        """Evaluate the log probability of point x."""

    @abstractmethod
    def _sample(self, key: jr.KeyArray, condition: Array | None = None) -> Array:
        """Sample a point from the distribution."""

    def _sample_and_log_prob(
        self, key: jr.KeyArray, condition: Array | None = None
    ) -> tuple[Array, Array]:
        """Sample a point from the distribution, and return its log probability.
        Subclasses can reimplement this method in cases where more efficient methods
        exists (e.g. for :class:`Transformed` distributions).
        """
        x = self._sample(key, condition)
        log_prob = self._log_prob(x, condition)
        return x, log_prob

    def log_prob(self, x: ArrayLike, condition: ArrayLike | None = None) -> Array:
        """Evaluate the log probability. Uses numpy like broadcasting if additional
        leading dimensions are passed.

        Args:
            x (ArrayLike): Points at which to evaluate density.
            condition (ArrayLike | None): Conditioning variables. Defaults to None.

        Returns:
            Array: Jax array of log probabilities.
        """
        x = self._argcheck_and_cast_x(x)
        condition = self._argcheck_and_cast_condition(condition)
        if condition is None:
            sig = _get_ufunc_signature([self.shape], [()])
            exclude = frozenset([1])
        else:
            sig = _get_ufunc_signature([self.shape, self.cond_shape], [()])
            exclude = frozenset()

        lps = jnp.vectorize(self._log_prob, signature=sig, excluded=exclude)(
            x, condition
        )
        return jnp.where(jnp.isnan(lps), -jnp.inf, lps)  # type: ignore

    def sample(
        self,
        key: jr.KeyArray,
        sample_shape: tuple[int, ...] = (),
        condition: ArrayLike | None = None,
    ) -> Array:
        """Sample from the distribution. For unconditional distributions, the output
        will be of shape ``sample_shape + dist.shape``. For conditional distributions,
        a batch dimension in the condition is supported, and the output shape will be
        ``sample_shape + condition_batch_shape + dist.shape``. See the example for more
        information.

        Args:
            key (jr.KeyArray): Jax random key.
            condition (ArrayLike | None): Conditioning variables. Defaults to None.
            sample_shape (tuple[int, ...]): Sample shape. Defaults to ().

        Example:
            The below example shows the behaviour of sampling, for an unconditional
            and a conditional distribution.

            .. testsetup::

                from flowjax.distributions import StandardNormal
                import jax.random as jr
                import jax.numpy as jnp
                from flowjax.flows import CouplingFlow
                from flowjax.bijections import Affine
                # For a unconditional distribution:
                key = jr.PRNGKey(0)
                dist = StandardNormal((2,))
                # For a conditional distribution
                cond_dist = CouplingFlow(
                    key, StandardNormal((2,)), cond_dim=3, transformer=Affine()
                    )

            For an unconditional distribution:

            .. doctest::

                >>> dist.shape
                (2,)
                >>> samples = dist.sample(key, (10, ))
                >>> samples.shape
                (10, 2)

            For a conditional distribution:

            .. doctest::

                >>> cond_dist.shape
                (2,)
                >>> cond_dist.cond_shape
                (3,)
                >>> # Sample 10 times for a particular condition
                >>> samples = cond_dist.sample(key, (10,), condition=jnp.ones(3))
                >>> samples.shape
                (10, 2)
                >>> # Sampling, batching over a condition
                >>> samples = cond_dist.sample(key, condition=jnp.ones((5, 3)))
                >>> samples.shape
                (5, 2)
                >>> # Sample 10 times for each of 5 conditioning variables
                >>> samples = cond_dist.sample(key, (10,), condition=jnp.ones((5, 3)))
                >>> samples.shape
                (10, 5, 2)


        """
        condition = self._argcheck_and_cast_condition(condition)
        excluded, signature = self._vectorize_sample_args()
        keys = self._get_sample_keys(key, sample_shape, condition)
        return jnp.vectorize(self._sample, excluded=excluded, signature=signature)(
            keys, condition
        )  # type: ignore

    def sample_and_log_prob(
        self,
        key: jr.KeyArray,
        sample_shape: tuple[int, ...] = (),
        condition: ArrayLike | None = None,
    ):
        """Sample the distribution and return the samples and corresponding log
        probabilities. For transformed distributions (especially flows), this will
        generally be more efficient than calling the methods seperately. Refer to the
        :py:meth:`~flowjax.distributions.Distribution.sample` documentation for more
        information.

        Args:
            key (jr.KeyArray): Jax random key.
            condition (ArrayLike | None): Conditioning variables. Defaults to None.
            sample_shape (tuple[int, ...]): Sample shape. Defaults to ().
        """
        condition = self._argcheck_and_cast_condition(condition)

        excluded, signature = self._vectorize_sample_args(sample_and_log_prob=True)
        keys = self._get_sample_keys(key, sample_shape, condition)

        return jnp.vectorize(
            self._sample_and_log_prob, excluded=excluded, signature=signature
        )(keys, condition)

    def _vectorize_sample_args(self, sample_and_log_prob=False):
        """Get the excluded arguments and ufunc signature for sample or
        sample_and_log_prob"""
        out_shapes = [self.shape, ()] if sample_and_log_prob else [self.shape]
        if self.cond_shape is None:
            excluded = frozenset([1])
            in_shapes = [(2,)]
        else:
            excluded = frozenset()
            in_shapes = [(2,), self.cond_shape]
        signature = _get_ufunc_signature(in_shapes, out_shapes)
        return excluded, signature

    def _get_sample_keys(self, key, sample_shape, condition):
        """Splits a key into an arrray of keys with shape
        sample_shape + leading_cond_shape + (2,)."""
        if self.cond_shape is None:
            key_shape = sample_shape
        else:
            leading_cond_shape = (
                condition.shape[: -len(self.cond_shape)]
                if len(self.cond_shape) > 0
                else condition.shape
            )
            key_shape = sample_shape + leading_cond_shape

        key_size = max(1, prod(key_shape))  # Still need 1 key for scalar sample
        keys = jnp.reshape(jr.split(key, key_size), key_shape + (2,))  # type: ignore
        return keys

    def _argcheck_and_cast_x(self, x) -> Array:
        x = arraylike_to_array(x, err_name="x")
        x_trailing = x.shape[-self.ndim :] if self.ndim > 0 else ()
        if x_trailing != self.shape:
            raise ValueError(
                "Expected trailing dimensions in x to match the distribution shape "
                f"{self.shape}; got x shape {x.shape}."
            )
        return x

    def _argcheck_and_cast_condition(self, condition) -> Array | None:
        if self.cond_shape is None:
            if condition is not None:
                raise ValueError(
                    "Expected condition to be None for unconditional distribution; "
                    f"got {condition}."
                )
            return None
        condition = arraylike_to_array(condition, err_name="condition")
        condition_trailing = (
            condition.shape[-len(self.cond_shape) :] if self.cond_ndim > 0 else ()
        )
        if condition_trailing != self.cond_shape:
            raise ValueError(
                "Expected trailing dimensions in condition to match cond_shape "
                f"{self.cond_shape}, but got condition shape {condition.shape}."
            )
        return condition

    @property
    def ndim(self):
        """The number of dimensions in the distribution (the length of the shape)."""
        return len(self.shape)

    @property
    def cond_ndim(self):
        """The number of dimensions of the conditioning variable (length of
        cond_shape)."""
        if self.cond_shape is not None:
            return len(self.cond_shape)
        return None


class Transformed(Distribution):
    """Form a distribution like object using a base distribution and a
    bijection. We take the forward bijection for use in sampling, and the inverse
    bijection for use in density evaluation.
    """

    base_dist: Distribution
    bijection: Bijection
    cond_shape: tuple[int, ...] | None

    def __init__(
        self,
        base_dist: Distribution,
        bijection: Bijection,
    ):
        """
        Args:
            base_dist (Distribution): Base distribution.
            bijection (Bijection): Bijection to transform distribution.

        Example:

        .. doctest::

            >>> from flowjax.distributions import StandardNormal, Transformed
            >>> from flowjax.bijections import Affine
            >>> normal = StandardNormal()
            >>> bijection = Affine(1)
            >>> transformed = Transformed(normal, bijection)


        .. warning::
            It is the currently the users responsibility to ensure the bijection is
            valid across the entire support of the distribution. Failure to do so may
            lead to to unexpected results. In future versions explicit constraints may
            be introduced.
        """
        self.base_dist = base_dist
        self.bijection = bijection
        self.shape = self.base_dist.shape
        self.cond_shape = merge_cond_shapes(
            (self.bijection.cond_shape, self.base_dist.cond_shape)
        )

    def _log_prob(self, x, condition=None):
        z, log_abs_det = self.bijection.inverse_and_log_det(x, condition)
        p_z = self.base_dist._log_prob(z, condition)  # pylint: disable W0212
        return p_z + log_abs_det

    def _sample(self, key, condition=None):
        base_sample = self.base_dist._sample(key, condition)
        return self.bijection.transform(base_sample, condition)

    def _sample_and_log_prob(self, key: jr.KeyArray, condition=None):
        # We overwrite the naive implementation of calling both methods seperately to
        # avoid computing the inverse transformation.
        base_sample, log_prob_base = self.base_dist._sample_and_log_prob(key, condition)
        sample, forward_log_dets = self.bijection.transform_and_log_det(
            base_sample, condition
        )
        return sample, log_prob_base - forward_log_dets


class StandardNormal(Distribution):
    """Implements a standard normal distribution, condition is ignored. Unlike
    :class:`Normal`, this has no trainable parameters.
    """

    def __init__(self, shape: tuple[int, ...] = ()):
        """
        Args:
            shape (tuple[int, ...]): The shape of the normal distribution. Defaults to
                ().
        """
        self.shape = shape
        self.cond_shape = None

    def _log_prob(self, x, condition=None):
        return jstats.norm.logpdf(x).sum()

    def _sample(self, key, condition=None):
        return jr.normal(key, self.shape)


class Normal(Transformed):
    """Implements an independent Normal distribution with mean and std for
    each dimension. `loc` and `scale` should be broadcastable.
    """

    bijection: Affine

    def __init__(self, loc: ArrayLike = 0, scale: ArrayLike = 1):
        """
        Args:
            loc (ArrayLike): Means. Defaults to 0.
            scale (ArrayLike): Standard deviations. Defaults to 1.
        """
        shape = jnp.broadcast_shapes(jnp.shape(loc), jnp.shape(scale))
        base_dist = StandardNormal(shape)
        bijection = Affine(loc=loc, scale=scale)
        super().__init__(base_dist, bijection)

    @property
    def loc(self):
        """Location of the distribution"""
        return self.bijection.loc

    @property
    def scale(self):
        """scale of the distribution"""
        return self.bijection.scale


class _StandardUniform(Distribution):
    r"""Implements a standard independent Uniform distribution, ie
    :math:`X \sim Uniform([0, 1]^d)`."""

    def __init__(self, shape: tuple[int, ...] = ()):
        self.shape = shape
        self.cond_shape = None

    def _log_prob(self, x, condition=None):
        return jstats.uniform.logpdf(x).sum()

    def _sample(self, key, condition=None):
        return jr.uniform(key, shape=self.shape)


class Uniform(Transformed):
    """Implements an independent uniform distribution between min and max for each
    dimension. `minval` and `maxval` should be broadcastable.
    """

    bijection: Affine

    def __init__(self, minval: ArrayLike, maxval: ArrayLike):
        """
        Args:
            minval (ArrayLike): Minimum values.
            maxval (ArrayLike): Maximum values.
        """
        minval, maxval = arraylike_to_array(minval), arraylike_to_array(maxval)
        shape = jnp.broadcast_shapes(minval.shape, maxval.shape)
        checkify.check(
            jnp.all(maxval >= minval),
            "Minimums must be less than the maximums.",
        )

        base_dist = _StandardUniform(shape)
        bijection = Affine(loc=minval, scale=maxval - minval)
        super().__init__(base_dist, bijection)

    @property
    def minval(self):
        """Minimum value of the uniform distribution."""
        return self.bijection.loc

    @property
    def maxval(self):
        """Maximum value of the uniform distribution."""
        return self.bijection.loc + self.bijection.scale


class _StandardGumbel(Distribution):
    """Standard gumbel distribution (https://en.wikipedia.org/wiki/Gumbel_distribution)."""

    def __init__(self, shape: tuple[int, ...] = ()):
        self.shape = shape
        self.cond_shape = None

    def _log_prob(self, x, condition=None):
        return -(x + jnp.exp(-x)).sum()

    def _sample(self, key, condition=None):
        return jr.gumbel(key, shape=self.shape)


class Gumbel(Transformed):
    """Gumbel distribution (https://en.wikipedia.org/wiki/Gumbel_distribution)"""

    bijection: Affine

    def __init__(self, loc: ArrayLike = 0, scale: ArrayLike = 1):
        """
        `loc` and `scale` should broadcast to the dimension of the distribution.

        Args:
            loc (ArrayLike): Location paramter.
            scale (ArrayLike): Scale parameter. Defaults to 1.0.
        """
        shape = jnp.broadcast_shapes(jnp.shape(loc), jnp.shape(scale))
        base_dist = _StandardGumbel(shape)
        bijection = Affine(loc, scale)

        super().__init__(base_dist, bijection)

    @property
    def loc(self):
        """Location of the distribution"""
        return self.bijection.loc

    @property
    def scale(self):
        """Scale of the distribution."""
        return self.bijection.scale


class _StandardCauchy(Distribution):
    """
    Implements standard cauchy distribution (loc=0, scale=1)
    Ref: https://en.wikipedia.org/wiki/Cauchy_distribution
    """

    def __init__(self, shape: tuple[int, ...] = ()):
        self.shape = shape
        self.cond_shape = None

    def _log_prob(self, x, condition=None):
        return jstats.cauchy.logpdf(x).sum()

    def _sample(self, key, condition=None):
        return jr.cauchy(key, shape=self.shape)


class Cauchy(Transformed):
    """Cauchy distribution (https://en.wikipedia.org/wiki/Cauchy_distribution)."""

    bijection: Affine

    def __init__(self, loc: ArrayLike = 0, scale: ArrayLike = 1):
        """
        `loc` and `scale` should broadcast to the dimension of the distribution.

        Args:
            loc (ArrayLike): Location paramter.
            scale (ArrayLike): Scale parameter. Defaults to 1.0.
        """
        shape = jnp.broadcast_shapes(jnp.shape(loc), jnp.shape(scale))
        base_dist = _StandardCauchy(shape)
        bijection = Affine(loc, scale)
        super().__init__(base_dist, bijection)

    @property
    def loc(self):
        """Location of the distribution"""
        return self.bijection.loc

    @property
    def scale(self):
        """scale of the distribution"""
        return self.bijection.scale


class _StandardStudentT(Distribution):
    """Implements student T distribution with specified degrees of freedom."""

    log_df: Array

    def __init__(self, df: ArrayLike):
        self.shape = jnp.shape(df)
        self.cond_shape = None
        self.log_df = jnp.log(df)

    def _log_prob(self, x, condition=None):
        return jstats.t.logpdf(x, df=self.df).sum()

    def _sample(self, key, condition=None):
        return jr.t(key, df=self.df, shape=self.shape)

    @property
    def df(self):
        """The degrees of freedom of the distibution."""
        return jnp.exp(self.log_df)


class StudentT(Transformed):
    """Student T distribution (https://en.wikipedia.org/wiki/Student%27s_t-distribution)."""

    bijection: Affine
    base_dist: _StandardStudentT

    def __init__(self, df: ArrayLike, loc: ArrayLike = 0, scale: ArrayLike = 1):
        """
        `df`, `loc` and `scale` broadcast to the dimension of the distribution.

        Args:
            df (ArrayLike): The degrees of freedom.
            loc (ArrayLike): Location parameter. Defaults to 0.0.
            scale (ArrayLike): Scale parameter. Defaults to 1.0.
        """
        df, loc, scale = jnp.broadcast_arrays(df, loc, scale)
        base_dist = _StandardStudentT(df)
        bijection = Affine(loc, scale)
        super().__init__(base_dist, bijection)

    @property
    def loc(self):
        """Location of the distribution"""
        return self.bijection.loc

    @property
    def scale(self):
        """scale of the distribution"""
        return self.bijection.scale

    @property
    def df(self):
        """The degrees of freedom of the distribution."""
        return self.base_dist.df


class SpecializeCondition(Distribution):
    """Utility class to specialise a conditional distribution to a particular instance
    of the conditioning variable. This makes the distribution act like an unconditional
    distribution, i.e. the distribution methods implicitly will use the condition passed
    on instantiation of the class.
    """

    def __init__(
        self,
        dist: Distribution,
        condition: ArrayLike,
        stop_gradient: bool = True,
    ):
        """
        Args:
            dist (Distribution): Conditional distribution to specialize.
            condition (ArrayLike, optional): Instance of conditioning variable with
                shape matching ``dist.cond_shape``. Defaults to None.
            stop_gradient (bool): Whether to use ``jax.lax.stop_gradient`` to prevent
                training of the condition array. Defaults to True.
        """
        condition = arraylike_to_array(condition)
        if self.dist.cond_shape != condition.shape:
            raise ValueError(
                f"Expected condition shape {self.dist.cond_shape}, got "
                f"{condition.shape}"
            )
        self.dist = dist
        self._condition = condition
        self.shape = dist.shape
        self.cond_shape = None
        self.stop_gradient = stop_gradient

    def _log_prob(self, x, condition=None):
        return self.dist._log_prob(x, self.condition)

    def _sample(self, key, condition=None):
        return self.dist._sample(key, self.condition)

    @property
    def condition(self):
        return stop_gradient(self._condition) if self.stop_gradient else self._condition



import equinox as eqx
import jax.numpy as jnp
import jax.tree_util as jtu
from jax import random

from flowjax.bijections import Affine
from flowjax.distributions import Normal, Transformed
from flowjax.train.data_fit import fit_to_data


def test_data_fit_filter_spec():
    dim = 3
    mean, std = jnp.ones(dim), jnp.ones(dim)
    base_dist = Normal(mean, std)
    flow = Transformed(base_dist, Affine(jnp.ones(dim), jnp.ones(dim)))

    # All params should change by default
    before = eqx.filter(flow, eqx.is_inexact_array)
    x = random.normal(random.PRNGKey(0), (100, dim))
    flow, _ = fit_to_data(random.PRNGKey(0), flow, x, max_epochs=1, batch_size=50)
    after = eqx.filter(flow, eqx.is_inexact_array)

    assert jnp.all(before.base_dist.bijection.loc != after.base_dist.bijection.loc)
    assert jnp.all(before.bijection.loc != after.bijection.loc)

    # We can provide a filter spec to avoid e.g. training the base distribution.
    before = eqx.filter(flow, eqx.is_inexact_array)
    filter_spec = jtu.tree_map(eqx.is_inexact_array, flow)
    filter_spec = eqx.tree_at(lambda tree: tree.base_dist, filter_spec, replace=False)
    flow, _ = fit_to_data(
        random.PRNGKey(0), flow, x, max_epochs=1, batch_size=50, filter_spec=filter_spec
    )
    after = eqx.filter(flow, eqx.is_inexact_array)

    assert jnp.all(before.base_dist.bijection.loc == after.base_dist.bijection.loc)
    assert jnp.all(before.bijection.loc != after.bijection.loc)



import jax
import jax.numpy as jnp
import pytest
from jax import random
from jax.scipy.linalg import block_diag

from flowjax.bijections.block_autoregressive_network import (
    BlockAutoregressiveNetwork,
    _block_tanh_activation,
)
from flowjax.masks import block_diag_mask


def test_BlockAutoregressiveNetwork():
    dim = 3
    x = jnp.ones(dim)
    key = random.PRNGKey(0)

    barn = BlockAutoregressiveNetwork(key, dim, None, depth=1, block_dim=4)
    y = barn.transform(x)
    assert y.shape == (dim,)
    auto_jacobian = jax.jacobian(barn.transform)(x)

    # Check autograd autoregressive
    assert jnp.all(jnp.triu(auto_jacobian, 1) == pytest.approx(0, abs=1e-7))
    assert jnp.all(jnp.diag(auto_jacobian) > 0)


def test_BlockAutoregressiveNetwork_conditioning():
    dim = 3
    cond_dim = 2
    x = jnp.ones(dim)
    key = random.PRNGKey(0)
    barn = BlockAutoregressiveNetwork(key, dim, cond_dim, depth=1, block_dim=4)
    y1 = barn.transform(x, jnp.ones(cond_dim))
    y2 = barn.transform(x, jnp.zeros(cond_dim))
    assert jnp.all(y1 != y2)


def test_block_tanh_activation():
    n_blocks = 2
    block_size = 3
    x = random.uniform(random.PRNGKey(0), (n_blocks * block_size,))
    tanh = _block_tanh_activation(n_blocks)

    y, log_det_3d = tanh(x)
    auto_jacobian = jax.jacobian(lambda a: tanh(a)[0])(x)
    mask = block_diag_mask((block_size, block_size), n_blocks)
    assert block_diag(*jnp.exp(log_det_3d)) == pytest.approx(
        auto_jacobian * mask, abs=1e-7
    )



"""Tanh bijection."""
import math

import jax
import jax.numpy as jnp

from flowjax.bijections.bijection import Bijection


def _tanh_log_grad(x):
    """log gradient vector of tanh transformation."""
    return -2 * (x + jax.nn.softplus(-2 * x) - jnp.log(2.0))


class Tanh(Bijection):
    """Tanh bijection."""

    def __init__(self, shape: tuple[int, ...] = ()) -> None:
        """
        Args:
            shape (tuple[int, ...] | None): Shape of the bijection. Defaults to None.
        """
        self.shape = shape
        self.cond_shape = None

    def transform(self, x, condition=None):
        x, _ = self._argcheck_and_cast(x)
        return jnp.tanh(x)

    def transform_and_log_det(self, x, condition=None):
        x, _ = self._argcheck_and_cast(x)
        return jnp.tanh(x), jnp.sum(_tanh_log_grad(x))

    def inverse(self, y, condition=None):
        y, _ = self._argcheck_and_cast(y)
        return jnp.arctanh(y)

    def inverse_and_log_det(self, y, condition=None):
        y, _ = self._argcheck_and_cast(y)
        x = jnp.arctanh(y)
        return x, -jnp.sum(_tanh_log_grad(x))


class TanhLinearTails(Bijection):
    """
    Tanh bijection, with linear "tails" beyond +/- max_val. Note due to the linear
    tails this does not guarantee the forward transformation will be constrained to
    [-1, 1]. This transform can be useful to "encourage" values to be within an interval
    (e.g. to subsequently apply some transformation defined on that interval), whilst
    avoiding issues with numerical instability.
    """

    max_val: float
    intercept: float
    linear_grad: float

    def __init__(self, max_val: float, shape: tuple[int, ...] = ()):
        """Create a tanh bijection with linear "tails" beyond +/- max_val.

        Args:
            max_val (int): Value above or below which the function becomes linear.
            shape (tuple[int, ...] | None): The shape of the bijection. Defaults to
                None.
        """
        self.max_val = max_val
        self.linear_grad = math.exp(_tanh_log_grad(max_val))
        self.intercept = math.tanh(max_val) - self.linear_grad * max_val
        self.shape = shape
        self.cond_shape = None

    def transform(self, x, condition=None):
        x, _ = self._argcheck_and_cast(x)
        is_linear = jnp.abs(x) >= self.max_val
        linear_y = self.linear_grad * x + jnp.sign(x) * self.intercept
        tanh_y = jnp.tanh(x)
        return jnp.where(is_linear, linear_y, tanh_y)

    def transform_and_log_det(self, x, condition=None):
        x, _ = self._argcheck_and_cast(x)
        y = self.transform(x)
        log_grads = jnp.where(
            jnp.abs(x) >= self.max_val, jnp.log(self.linear_grad), _tanh_log_grad(x)
        )
        return y, jnp.sum(log_grads)  # type: ignore

    def inverse(self, y, condition=None):
        y, _ = self._argcheck_and_cast(y)
        is_linear = jnp.abs(y) >= jnp.tanh(self.max_val)
        x_linear = (y - jnp.sign(y) * self.intercept) / self.linear_grad
        x_arctan = jnp.arctanh(y)
        return jnp.where(is_linear, x_linear, x_arctan)

    def inverse_and_log_det(self, y, condition=None):
        y, _ = self._argcheck_and_cast(y)
        x = self.inverse(y)
        log_grads = jnp.where(
            jnp.abs(y) >= jnp.tanh(self.max_val),
            jnp.log(self.linear_grad),
            _tanh_log_grad(x),
        )
        return x, -jnp.sum(log_grads)  # type: ignore



"""Block autoregressive neural network components."""
from typing import Callable

import equinox as eqx
import jax
import jax.numpy as jnp
from jax import Array, random
from jax.nn.initializers import glorot_uniform
from jax.random import KeyArray

from flowjax.bijections.tanh import Tanh
from flowjax.masks import block_diag_mask, block_tril_mask


def _tanh_log_grad(x):
    """log gradient vector of tanh transformation."""
    return -2 * (x + jax.nn.softplus(-2 * x) - jnp.log(2.0))


class BlockAutoregressiveLinear(eqx.Module):
    """Block autoregressive neural network layer (https://arxiv.org/abs/1904.04676).
    Conditioning variables are incorporated by appending columns (one for each
    conditioning variable) to the left of the block diagonal weight matrix.
    """

    n_blocks: int
    block_shape: tuple
    cond_dim: int | None
    weights: Array
    bias: Array
    weight_log_scale: Array
    in_features: int
    out_features: int
    b_diag_mask: Array
    b_diag_mask_idxs: Array
    b_tril_mask: Array

    def __init__(
        self,
        key: KeyArray,
        n_blocks: int,
        block_shape: tuple,
        cond_dim: int | None = None,
        init: Callable | None = None,
    ):
        """
        Args:
            key KeyArray: Random key
            n_blocks (int): Number of diagonal blocks (dimension of original input).
            block_shape (tuple): The shape of the (unconstrained) blocks.
            cond_dim (int | None): Number of additional conditioning variables.
                Defaults to None.
            init (Callable | None): Default initialisation method for the weight
                matrix. Defaults to ``glorot_uniform()``.
        """
        init = init if init is not None else glorot_uniform()
        self.cond_dim = cond_dim

        if cond_dim is None:
            cond_dim = 0

        cond_size = (block_shape[0] * n_blocks, cond_dim)

        self.b_diag_mask = jnp.column_stack(
            (block_diag_mask(block_shape, n_blocks), jnp.zeros(cond_size, jnp.int32))
        )

        self.b_tril_mask = jnp.column_stack(
            (block_tril_mask(block_shape, n_blocks), jnp.ones(cond_size, jnp.int32))
        )

        self.b_diag_mask_idxs = jnp.where(
            self.b_diag_mask, size=block_shape[0] * block_shape[1] * n_blocks
        )  # type: ignore

        in_features, out_features = (
            block_shape[1] * n_blocks + cond_dim,
            block_shape[0] * n_blocks,
        )

        *w_key, bias_key, scale_key = random.split(key, n_blocks + 2)

        self.weights = init(w_key[0], (out_features, in_features)) * (
            self.b_tril_mask + self.b_diag_mask
        )
        self.bias = (random.uniform(bias_key, (out_features,)) - 0.5) * (
            2 / jnp.sqrt(out_features)
        )

        self.n_blocks = n_blocks
        self.block_shape = block_shape
        self.cond_dim = cond_dim
        self.weight_log_scale = jnp.log(random.uniform(scale_key, (out_features, 1)))
        self.in_features = in_features
        self.out_features = out_features

    def get_normalised_weights(self):
        """Carries out weight normalisation."""
        weights = (
            jnp.exp(self.weights) * self.b_diag_mask + self.weights * self.b_tril_mask
        )
        weight_norms = jnp.linalg.norm(weights, axis=-1, keepdims=True)
        return jnp.exp(self.weight_log_scale) * weights / weight_norms

    def __call__(self, x, condition=None):
        """returns output y, and components of weight matrix needed log_det component
        (n_blocks, block_shape[0], block_shape[1]).
        """
        weights = self.get_normalised_weights()
        if condition is not None:
            x = jnp.concatenate((x, condition))
        y = weights @ x + self.bias
        jac_3d = weights[self.b_diag_mask_idxs].reshape(
            self.n_blocks, *self.block_shape
        )
        return y, jnp.log(jac_3d)


def _block_tanh_activation(n_blocks: int):
    """Returms a Tanh activation compatible with a block neural autoregressive network,
    i.e. returning the transformed variable and the absolute log gradients as a 3D
    array with shape (n_blocks,) + block_shape
    """

    def activation(x: Array):
        tanh = Tanh()
        y, log_jacobian = jax.vmap(tanh.transform_and_log_det)(x)
        return y, _reshape_jacobian_to_3d(log_jacobian, n_blocks)

    return activation


def _reshape_jacobian_to_3d(vals, n_blocks):
    block_dim = vals.shape[0] // n_blocks
    log_det = jnp.full((n_blocks, block_dim, block_dim), -jnp.inf)
    log_det = log_det.at[:, jnp.arange(block_dim), jnp.arange(block_dim)].set(
        vals.reshape(n_blocks, block_dim)
    )
    return log_det



"""Common loss functions for training normalizing flows. The loss functions are
callables, with the first two arguments being the partitioned distribution (see
equinox.partition)."""
from typing import Callable

import equinox as eqx
import jax.numpy as jnp
import jax.random as jr
from jax import Array, vmap
from jax.lax import stop_gradient
from jax.scipy.special import logsumexp
from jax.typing import ArrayLike

from flowjax.distributions import Distribution


class MaximumLikelihoodLoss:
    """Loss for fitting a flow with maximum likelihood (negative log likelihood). Can
    be used to learn either conditional or unconditional distributions.
    """

    @eqx.filter_jit
    def __call__(
        self,
        static: Distribution,
        params: Distribution,
        x: Array,
        condition: Array | None = None,
    ):
        dist = eqx.combine(static, params)
        return -dist.log_prob(x, condition).mean()


class ContrastiveLoss:
    r"""Loss function for use in a sequential neural posterior estimation algorithm.
    Learns a posterior ``p(x|condition)``. Contrastive samples for each ``x`` are
    generated from other x samples in the batch.

    Note, that in a simulation based inference context, often :math:`x` is used to
    denote simulations, and :math:`\theta` for simulation parameters. However, for
    consistency with the rest of the package, we use ``x`` to represent the target
    variable (the simulator parameters), and ``condition`` for the conditioning variable
    (the simulator output/oberved data).

    References:
        - https://arxiv.org/abs/1905.07488
        - https://arxiv.org/abs/2002.03712

    """

    def __init__(self, prior: Distribution, n_contrastive: int):
        """
        Args:
            prior (Distribution): The prior distribution over x (the target variable).
            n_contrastive (int): The number of contrastive samples/atoms to use when
                computing the loss.
        """
        self.prior = prior
        self.n_contrastive = n_contrastive

    @eqx.filter_jit
    def __call__(
        self,
        static: Distribution,
        params: Distribution,
        x: Array,
        condition: Array | None = None,
    ):
        dist = eqx.combine(params, static)
        contrastive = self._get_contrastive(x)
        joint_log_odds = dist.log_prob(x, condition) - self.prior.log_prob(x)
        contrastive_log_odds = dist.log_prob(
            contrastive, condition
        ) - self.prior.log_prob(contrastive)
        contrastive_log_odds = jnp.clip(contrastive_log_odds, -5)  # Clip for stability
        return -(joint_log_odds - logsumexp(contrastive_log_odds, axis=0)).mean()

    def _get_contrastive(self, theta):
        if theta.shape[0] <= self.n_contrastive:
            raise ValueError(
                f"Number of contrastive samples {self.n_contrastive} must be less than "
                f"the size of theta {theta.shape}."
            )
        # Rolling window over theta batch to create contrastive samples.
        idx = jnp.arange(len(theta))[:, None] + jnp.arange(self.n_contrastive)[None, :]
        contrastive = jnp.roll(theta[idx], -1, axis=0)  # Ensure mismatch with condition
        contrastive = jnp.swapaxes(contrastive, 0, 1)  # (contrastive, batch_size, dim)
        return contrastive


class ElboLoss:
    """The negative evidence lower bound (ELBO), approximated using samples."""

    target: Callable[[ArrayLike], Array]
    num_samples: int
    stick_the_landing: bool

    def __init__(
        self,
        target: Callable[[ArrayLike], Array],
        num_samples: int,
        stick_the_landing: bool = False,
    ):
        """
        Args:
            num_samples (int): Number of samples to use in the ELBO approximation.
            target (Callable[[ArrayLike], Array]): The target, i.e. log posterior
                density up to an additive constant / the negative of the potential
                function, evaluated for a single point.
            stick_the_landing (bool): Whether to use the (often) lower variance ELBO
                gradient estimator introduced in https://arxiv.org/pdf/1703.09194.pdf.
                Note for flows this requires evaluating the flow in both directions
                (running the forward and inverse transformation). For some flow
                architectures, this may be computationally expensive due to assymetrical
                computational complexity between the forward and inverse transformation.
                Defaults to False.
        """
        self.target = target
        self.num_samples = num_samples
        self.stick_the_landing = stick_the_landing

    @eqx.filter_jit
    def __call__(self, params: Distribution, static: Distribution, key: jr.KeyArray):
        dist = eqx.combine(params, static)

        if self.stick_the_landing:
            # Requries both forward and inverse pass
            samples = dist.sample(key, (self.num_samples,))
            dist = eqx.combine(stop_gradient(params), static)
            log_probs = dist.log_prob(samples)

        else:
            # Requires only forward pass through the flow.
            samples, log_probs = dist.sample_and_log_prob(key, (self.num_samples,))

        target_density = vmap(self.target)(samples)
        return (log_probs - target_density).mean()



"""Exponential bijection"""
import jax.numpy as jnp

from flowjax.bijections.bijection import Bijection


class Exp(Bijection):
    """Elementwise exponential transform (forward) and log transform (inverse)."""

    def __init__(self, shape: tuple[int, ...] = ()):
        """
        Args:
            shape (tuple[int, ...] | None): Shape of the bijection.
                Defaults to None.
        """
        self.shape = shape
        self.cond_shape = None

    def transform(self, x, condition=None):
        x, _ = self._argcheck_and_cast(x)
        return jnp.exp(x)

    def transform_and_log_det(self, x, condition=None):
        x, _ = self._argcheck_and_cast(x)
        return jnp.exp(x), x.sum()

    def inverse(self, y, condition=None):
        y, _ = self._argcheck_and_cast(y)
        return jnp.log(y)

    def inverse_and_log_det(self, y, condition=None):
        y, _ = self._argcheck_and_cast(y)
        x = jnp.log(y)
        return x, -x.sum()



"""The train sub-package contains basic functions for training flows, to
samples from a target distribution, or by using variational inference.
"""

from .data_fit import fit_to_data
from .variational_fit import fit_to_variational_target

__all__ = [
    "fit_to_data",
    "fit_to_variational_target",
]



"Tests for train_utils.py"
import jax.numpy as jnp
import jax.random as jr
import pytest

from flowjax.train.train_utils import count_fruitless, get_batches, train_val_split


def test_train_val_split():
    key = jr.PRNGKey(0)
    arrays = [jnp.arange(10).reshape(5, 2) for _ in range(3)]

    train, val = train_val_split(key, arrays, val_prop=0.2)

    # Check shapes
    assert all(train_arr.shape == (4, 2) for train_arr in train)
    assert all(val_arr.shape == (1, 2) for val_arr in val)

    arrays = [jnp.ones((5, 2)), jnp.ones((3, 5))]

    # Axis mismatch
    with pytest.raises(ValueError):
        train_val_split(key, arrays, val_prop=0.2)


def test_count_fruitless():
    assert count_fruitless([12, 2, 3, 4]) == 2
    assert count_fruitless([0]) == 0
    assert count_fruitless([0, 12]) == 1


def test_get_batches():
    arrays = [jnp.arange(26).reshape(13, 2)] * 2
    out = get_batches(arrays, batch_size=4)
    assert out[0].shape == (3, 4, 2)



"""Basic training script for fitting a flow using variational inference."""
from typing import Any, Callable

import equinox as eqx
import jax.random as jr
import optax
from tqdm import tqdm

from flowjax.distributions import Distribution
from flowjax.train.train_utils import step

PyTree = Any


def fit_to_variational_target(
    key: jr.KeyArray,
    dist: Distribution,
    loss_fn: Callable,
    steps: int = 100,
    learning_rate: float = 5e-4,
    optimizer: optax.GradientTransformation | None = None,
    filter_spec: Callable | PyTree = eqx.is_inexact_array,
    show_progress: bool = True,
):
    """Train a distribution (e.g. a flow) by variational inference to a target
    (e.g. an unnormalized density).

    Args:
        key (jr.KeyArray): Jax PRNGKey.
        dist (Distribution): Distribution object, trainable parameters are found
            using equinox.is_inexact_array.
        loss_fn (Callable | None): The loss function to optimize (e.g. the ElboLoss).
        steps (int, optional): The number of training steps to run. Defaults to 100.
        learning_rate (float, optional): Learning rate. Defaults to 5e-4.
        optimizer (optax.GradientTransformation | None, optional): Optax optimizer. If
            provided, this overrides the default Adam optimizer, and the learning_rate
            is ignored. Defaults to None.
        filter_spec (Callable | PyTree, optional): Equinox `filter_spec` for
            specifying trainable parameters. Either a callable `leaf -> bool`, or a
            PyTree with prefix structure matching `dist` with True/False values.
            Defaults to eqx.is_inexact_array.
        show_progress (bool, optional): Whether to show progress bar. Defaults to True.

    Returns:
        tuple: (distribution, losses).
    """
    if optimizer is None:
        optimizer = optax.adam(learning_rate)

    @eqx.filter_value_and_grad
    def loss_val_and_grad(trainable, static, key):
        model = eqx.combine(trainable, static)
        return loss_fn(model, key)

    params, static = eqx.partition(dist, filter_spec)
    opt_state = optimizer.init(params)

    losses = []

    best_params = params
    keys = tqdm(jr.split(key, steps), disable=not show_progress)

    for key in keys:
        params, opt_state, loss = step(
            optimizer, opt_state, loss_fn, params, static, key
        )
        losses.append(loss.item())
        keys.set_postfix({"loss": loss.item()})
        if loss.item() == min(losses):
            best_params = params

    return eqx.combine(best_params, static), losses



import equinox as eqx
import jax.numpy as jnp
import jax.random as jr
import pytest
from jax import tree_map

from flowjax.bijections import RationalQuadraticSpline


def test_RationalQuadraticSpline_tails():
    key = jr.PRNGKey(0)
    x = jnp.array([-20, 0.1, 2, 20])
    spline = RationalQuadraticSpline(knots=10, interval=3, shape=x.shape)

    # Change to random initialisation, rather than identity.
    spline = tree_map(
        lambda x: jr.normal(key, x.shape) if eqx.is_inexact_array(x) else x, spline
    )

    y = spline.transform(x)
    expected_changed = jnp.array([True, False, False, True])  # identity padding
    assert ((jnp.abs((y - x)) <= 1e-5) == expected_changed).all()


def test_RationalQuadraticSpline_init():
    # Test it is initialised at the identity
    x = jnp.array([-1, 0.1, 2, 1])
    key = jr.PRNGKey(0)
    spline = RationalQuadraticSpline(knots=10, interval=3, shape=x.shape)
    y = spline.transform(x)
    assert pytest.approx(x, abs=1e-6) == y

    shape = spline.bijection.unbounded_derivatives.shape
    spline = eqx.tree_at(
        lambda b: b.bijection.unbounded_derivatives, spline, jr.normal(key, shape)
    )
    y = spline.transform(x)
    assert pytest.approx(x, abs=1e-6) != y



"""Function to fit flows to samples from a distribution."""
from typing import Any, Callable

import equinox as eqx
import jax.numpy as jnp
import jax.random as jr
import optax
from jax.typing import ArrayLike
from tqdm import tqdm

from flowjax.distributions import Distribution
from flowjax.train.losses import MaximumLikelihoodLoss
from flowjax.train.train_utils import (
    count_fruitless,
    get_batches,
    step,
    train_val_split,
)

PyTree = Any


def fit_to_data(
    key: jr.KeyArray,
    dist: Distribution,
    x: ArrayLike,
    condition: ArrayLike = None,
    loss_fn: Callable | None = None,
    max_epochs: int = 100,
    max_patience: int = 5,
    batch_size: int = 100,
    val_prop: float = 0.1,
    learning_rate: float = 5e-4,
    optimizer: optax.GradientTransformation | None = None,
    filter_spec: Callable | PyTree = eqx.is_inexact_array,
    show_progress: bool = True,
):
    """Train a distribution (e.g. a flow) to samples from the target distribution p(x)
    or p(x|condition). Note that the last batch in each epoch is dropped if truncated.

    Args:
        key (KeyArray): Jax random seed.
        dist (Distribution): Distribution object.
        x (ArrayLike): Samples from target distribution.
        condition (ArrayLike | None): Conditioning variables. Defaults to None.
        loss_fn (Callable | None): Loss function. Defaults to MaximumLikelihoodLoss.
        max_epochs (int): Maximum number of epochs. Defaults to 100.
        max_patience (int): Number of consecutive epochs with no validation
            loss improvement after which training is terminated. Defaults to 5.
        batch_size (int): Batch size. Defaults to 100.
        val_prop (float): Proportion of data to use in validation set. Defaults to 0.1.
        learning_rate (float): Adam learning rate. Defaults to 5e-4.
        optimizer (optax.GradientTransformation): Optax optimizer. If provided, this
            overrides the default Adam optimizer, and the learning_rate is ignored.
            Defaults to None.
        filter_spec (Callable | PyTree): Equinox `filter_spec` for specifying trainable
            parameters. Either a callable `leaf -> bool`, or a PyTree with prefix
            structure matching `dist` with True/False values. Defaults to
            `eqx.is_inexact_array`.
        show_progress (bool): Whether to show progress bar. Defaults to True.
    """
    data = (x,) if condition is None else (x, condition)
    data = tuple(jnp.asarray(a) for a in data)

    if optimizer is None:
        optimizer = optax.adam(learning_rate)

    if loss_fn is None:
        loss_fn = MaximumLikelihoodLoss()

    params, static = eqx.partition(dist, filter_spec)
    best_params = params
    opt_state = optimizer.init(params)

    # train val split
    key, subkey = jr.split(key)
    train_data, val_data = train_val_split(key, data, val_prop=val_prop)
    losses = {"train": [], "val": []}

    loop = tqdm(range(max_epochs), disable=not show_progress)

    for _ in loop:
        # Shuffle data
        key, *subkeys = jr.split(key, 3)
        train_data = [jr.permutation(subkeys[0], a) for a in train_data]
        val_data = [jr.permutation(subkeys[1], a) for a in val_data]

        # Train epoch
        batch_losses = []
        for batch in zip(*get_batches(train_data, batch_size), strict=True):
            key, subkey = jr.split(key)
            params, opt_state, loss_i = step(
                optimizer, opt_state, loss_fn, params, static, *batch
            )
            batch_losses.append(loss_i)
        losses["train"].append(sum(batch_losses) / len(batch_losses))

        # Val epoch
        batch_losses = []
        for batch in zip(*get_batches(val_data, batch_size), strict=True):
            loss_i = loss_fn(params, static, *batch)
            batch_losses.append(loss_i)
        losses["val"].append(sum(batch_losses) / len(batch_losses))

        loop.set_postfix({k: v[-1] for k, v in losses.items()})
        if losses["val"][-1] == min(losses["val"]):
            best_params = params

        elif count_fruitless(losses["val"]) > max_patience:
            loop.set_postfix_str(f"{loop.postfix} (Max patience reached)")
            break

    dist = eqx.combine(best_params, static)
    return dist, losses



"General tests for bijections (including transformers)."
import equinox as eqx
import jax
import jax.numpy as jnp
import jax.random as jr
import pytest

from flowjax.bijections import (
    AdditiveCondition,
    Affine,
    BlockAutoregressiveNetwork,
    Chain,
    Concatenate,
    Coupling,
    EmbedCondition,
    Exp,
    Flip,
    MaskedAutoregressive,
    Partial,
    Permute,
    Planar,
    RationalQuadraticSpline,
    Scan,
    SoftPlus,
    Stack,
    Tanh,
    TanhLinearTails,
    TriangularAffine,
)

DIM = 5
COND_DIM = 2
KEY = jr.PRNGKey(0)
POS_DEF_TRAINGLES = jnp.full((DIM, DIM), 0.5) + jnp.diag(jnp.ones(DIM))


def get_maf_layer(key):
    """Get a masked autoregressive flow layer."""
    return MaskedAutoregressive(
        key, Affine(), DIM, cond_dim=COND_DIM, nn_width=5, nn_depth=5
    )


bijections = {
    "Flip": Flip((DIM,)),
    "Permute": Permute(jnp.flip(jnp.arange(DIM))),
    "Permute (3D)": Permute(
        jnp.reshape(jr.permutation(KEY, jnp.arange(2 * 3 * 4)), (2, 3, 4))
    ),
    "Partial (int)": Partial(Affine(jnp.array(2), jnp.array(2)), 0, (DIM,)),
    "Partial (bool array)": Partial(
        Flip((3,)), jnp.array([True, False, True, False, True]), (DIM,)
    ),
    "Partial (int array)": Partial(Flip((2,)), jnp.array([0, 4]), (DIM,)),
    "Partial (slice)": Partial(Affine(jnp.zeros(3)), slice(0, 3), (DIM,)),
    "Affine": Affine(jnp.ones(DIM), jnp.full(DIM, 2)),
    "Tanh": Tanh((DIM,)),
    "Exp": Exp((DIM,)),
    "SoftPlus": SoftPlus((DIM,)),
    "TanhLinearTails": TanhLinearTails(1, (DIM,)),
    "TriangularAffine (lower)": TriangularAffine(jnp.arange(DIM), POS_DEF_TRAINGLES),
    "TriangularAffine (upper)": TriangularAffine(
        jnp.arange(DIM), POS_DEF_TRAINGLES, lower=False
    ),
    "TriangularAffine (weight_norm)": TriangularAffine(
        jnp.arange(DIM), POS_DEF_TRAINGLES, weight_normalisation=True
    ),
    "RationalQuadraticSpline": RationalQuadraticSpline(knots=4, interval=1, shape=(5,)),
    "Coupling (unconditional)": Coupling(
        KEY,
        Affine(),
        untransformed_dim=DIM // 2,
        dim=DIM,
        cond_dim=None,
        nn_width=10,
        nn_depth=2,
    ),
    "Coupling (conditional)": Coupling(
        KEY,
        Affine(),
        untransformed_dim=DIM // 2,
        dim=DIM,
        cond_dim=COND_DIM,
        nn_width=10,
        nn_depth=2,
    ),
    "MaskedAutoregressive_Affine (unconditional)": MaskedAutoregressive(
        KEY, Affine(), cond_dim=0, dim=DIM, nn_width=10, nn_depth=2
    ),
    "MaskedAutoregressive_Affine (conditional)": MaskedAutoregressive(
        KEY, Affine(), cond_dim=COND_DIM, dim=DIM, nn_width=10, nn_depth=2
    ),
    "MaskedAutoregressiveRationalQuadraticSpline (unconditional)": MaskedAutoregressive(
        KEY,
        RationalQuadraticSpline(5, 3),
        dim=DIM,
        cond_dim=0,
        nn_width=10,
        nn_depth=2,
    ),
    "BlockAutoregressiveNetwork (unconditional)": BlockAutoregressiveNetwork(
        KEY, dim=DIM, cond_dim=0, block_dim=3, depth=1
    ),
    "BlockAutoregressiveNetwork (conditional)": BlockAutoregressiveNetwork(
        KEY, dim=DIM, cond_dim=COND_DIM, block_dim=3, depth=1
    ),
    "AdditiveCondtition": AdditiveCondition(
        lambda condition: jnp.arange(DIM) * jnp.sum(condition), (DIM,), (COND_DIM,)
    ),
    "EmbedCondition": EmbedCondition(
        BlockAutoregressiveNetwork(KEY, dim=DIM, cond_dim=1, block_dim=3, depth=1),
        eqx.nn.MLP(2, 1, 3, 1, key=KEY),
        (COND_DIM,),  # Raw
    ),
    "Chain": Chain([Flip((DIM,)), Affine(jnp.ones(DIM), jnp.full(DIM, 2))]),
    "Scan": Scan(eqx.filter_vmap(get_maf_layer)(jr.split(KEY, 3))),
    "Concatenate": Concatenate([Affine(jnp.ones(3)), Tanh(shape=(3,))]),
    "ConcatenateAxis1": Concatenate(
        [Affine(jnp.ones((3, 3))), Tanh(shape=((3, 3)))], axis=1
    ),
    "ConcatenateAxis-1": Concatenate(
        [Affine(jnp.ones((3, 3))), Tanh(shape=((3, 3)))], axis=-1
    ),
    "Stack": Stack([Tanh(()), Affine(), Tanh(())]),
    "StackAxis1": Stack([Tanh((2,)), Affine(jnp.ones(2)), Tanh((2,))], axis=1),
    "StackAxis-1": Stack(
        [Affine(jr.uniform(k, (1, 2, 3))) for k in jr.split(KEY, 3)], axis=-1
    ),
    "Planar": Planar(
        KEY,
        DIM,
    ),
}


@pytest.mark.parametrize("bijection", bijections.values(), ids=bijections.keys())
def test_transform_inverse(bijection):
    """Tests transform and inverse methods."""
    shape = bijection.shape if bijection.shape is not None else (DIM,)
    x = jr.normal(jr.PRNGKey(0), shape)
    if bijection.cond_shape is not None:
        cond = jr.normal(jr.PRNGKey(0), bijection.cond_shape)
    else:
        cond = None
    y = bijection.transform(x, cond)
    try:
        x_reconstructed = bijection.inverse(y, cond)
        assert x == pytest.approx(x_reconstructed, abs=1e-4)
    except NotImplementedError:
        pass


@pytest.mark.parametrize("bijection", bijections.values(), ids=bijections.keys())
def test_transform_inverse_and_log_dets(bijection):
    """Tests the transform_and_log_det and inverse_and_log_det methods,
    by 1) checking invertibility and 2) comparing log dets to those obtained with
    automatic differentiation."""
    shape = bijection.shape if bijection.shape is not None else (DIM,)
    x = jr.normal(jr.PRNGKey(0), shape)

    if bijection.cond_shape is not None:
        cond = jr.normal(jr.PRNGKey(0), bijection.cond_shape)
    else:
        cond = None

    # We flatten the function so auto_jacobian is calculated correctly
    def flat_transform(x_flat, cond):
        x = x_flat.reshape(shape)
        y = bijection.transform(x, cond)
        return y.ravel()

    auto_jacobian = jax.jacobian(flat_transform)(x.ravel(), cond)
    auto_log_det = jnp.log(jnp.abs(jnp.linalg.det(auto_jacobian)))
    y, logdet = bijection.transform_and_log_det(x, cond)
    assert logdet == pytest.approx(auto_log_det, abs=1e-4)

    try:
        x_reconstructed, logdetinv = bijection.inverse_and_log_det(y, cond)
        assert logdetinv == pytest.approx(-auto_log_det, abs=1e-4)
        assert x == pytest.approx(x_reconstructed, abs=1e-4)

    except NotImplementedError:
        pass



import jax
import jax.numpy as jnp
import pytest
from jax import random
from jax.scipy.linalg import block_diag

from flowjax.masks import rank_based_mask
from flowjax.nn import AutoregressiveMLP, BlockAutoregressiveLinear, MaskedLinear


def test_BlockAutoregressiveLinear():
    block_shape = (3, 2)
    layer = BlockAutoregressiveLinear(jax.random.PRNGKey(0), 3, block_shape)
    x = jnp.ones(6)
    a, log_jac_3d = layer(x)
    assert log_jac_3d.shape == (3, *block_shape)

    # Check block diag log jacobian matches autodif.
    auto_jacobian = jax.jacobian(lambda x: layer(x)[0])(x) * layer.b_diag_mask
    assert block_diag(*jnp.exp(log_jac_3d)) == pytest.approx(auto_jacobian, abs=1e-7)


def test_MaskedLinear():
    key = random.PRNGKey(0)
    in_ranks = jnp.arange(3)
    out_ranks = jnp.arange(5).repeat(2)
    mask = rank_based_mask(in_ranks, out_ranks)
    layer = MaskedLinear(mask, key=key)
    x = jnp.ones(len(in_ranks))

    # Ensure dependency structure correctly enforced
    non_zero_jac = jnp.abs(jax.jacobian(layer)(x)) > 1e-8
    expected = rank_based_mask(in_ranks, out_ranks)
    assert jnp.all(expected == non_zero_jac)


def test_AutoregressiveMLP():
    key = random.PRNGKey(0)
    in_size = 4
    in_ranks = jnp.arange(in_size)
    hidden_ranks = jnp.arange(6) % in_size
    out_ranks = jnp.arange(in_size).repeat(2)
    mlp = AutoregressiveMLP(in_ranks, hidden_ranks, out_ranks, depth=3, key=key)
    x = jnp.ones(in_size)
    y = mlp(x)
    assert y.shape == out_ranks.shape

    # Check connectivity matrix matches expected dependency structure
    a = mlp.layers[-1].mask
    for layer in reversed(mlp.layers[:-1]):
        a = a @ layer.mask

    expected = rank_based_mask(jnp.arange(in_size), out_ranks)
    assert jnp.all((a > 0) == expected)



"""nn package contains neural network architectures of use for flows (invertible
neural networks).
"""

from .block_autoregressive import BlockAutoregressiveLinear, _block_tanh_activation
from .masked_autoregressive import AutoregressiveMLP, MaskedLinear

__all__ = [
    "MaskedLinear",
    "AutoregressiveMLP",
    "BlockAutoregressiveLinear",
    "_block_tanh_activation",
]



"""Masked autoregressive network and bijection."""

from functools import partial
from typing import Callable

import equinox as eqx
import jax
import jax.nn as jnn
import jax.numpy as jnp
from jax import Array
from jax.random import KeyArray

from flowjax.bijections.bijection import Bijection
from flowjax.bijections.jax_transforms import Batch
from flowjax.nn import AutoregressiveMLP
from flowjax.utils import get_ravelled_bijection_constructor


class MaskedAutoregressive(Bijection):
    """Masked autoregressive bijection implementation (https://arxiv.org/abs/1705.07057v4).
    The transformer is parameterised by a neural network, with weights masked to ensure
    an autoregressive structure.
    """

    transformer_constructor: Callable
    autoregressive_mlp: AutoregressiveMLP

    def __init__(
        self,
        key: KeyArray,
        transformer: Bijection,
        dim: int,
        cond_dim: int | None,
        nn_width: int,
        nn_depth: int,
        nn_activation: Callable = jnn.relu,
    ) -> None:
        """
        Args:
            key (KeyArray): Jax PRNGKey
            transformer (Bijection): Bijection with shape () to be parameterised by the
                autoregressive network.
            dim (int): Dimension.
            cond_dim (int | None): Dimension of any conditioning variables.
            nn_width (int): Neural network width.
            nn_depth (int): Neural network depth.
            nn_activation (Callable): Neural network activation. Defaults to jnn.relu.
        """
        if transformer.shape != () or transformer.cond_shape is not None:
            raise ValueError(
                "Only unconditional transformers with shape () are supported."
            )

        constructor, transformer_init_params = get_ravelled_bijection_constructor(
            transformer
        )

        if cond_dim is None:
            self.cond_shape = None
            in_ranks = jnp.arange(dim)
        else:
            self.cond_shape = (cond_dim,)
            # we give conditioning variables rank -1 (no masking of edges to output)
            in_ranks = jnp.hstack((jnp.arange(dim), -jnp.ones(cond_dim)))

        hidden_ranks = jnp.arange(nn_width) % dim
        out_ranks = jnp.repeat(jnp.arange(dim), transformer_init_params.size)

        autoregressive_mlp = AutoregressiveMLP(
            in_ranks,
            hidden_ranks,
            out_ranks,
            nn_depth,
            nn_activation,
            key=key,
        )

        # Initialise bias terms to match the provided transformer parameters
        self.autoregressive_mlp = eqx.tree_at(
            where=lambda t: t.layers[-1].linear.bias,  # type: ignore
            pytree=autoregressive_mlp,
            replace=jnp.tile(transformer_init_params, dim),
        )

        self.transformer_constructor = constructor
        self.shape = (dim,)
        self.cond_shape = None if cond_dim is None else (cond_dim,)

    def transform(self, x, condition=None):
        nn_input = x if condition is None else jnp.hstack((x, condition))
        transformer_params = self.autoregressive_mlp(nn_input)
        transformer = self._flat_params_to_transformer(transformer_params)
        return transformer.transform(x)

    def transform_and_log_det(self, x, condition=None):
        nn_input = x if condition is None else jnp.hstack((x, condition))
        transformer_params = self.autoregressive_mlp(nn_input)
        transformer = self._flat_params_to_transformer(transformer_params)
        return transformer.transform_and_log_det(x)

    def inverse(self, y, condition=None):
        init = (y, 0)
        fn = partial(self.inv_scan_fn, condition=condition)
        (x, _), _ = jax.lax.scan(fn, init, None, length=len(y))
        return x

    def inv_scan_fn(self, init, _, condition):
        """One 'step' in computing the inverse."""
        y, rank = init
        nn_input = y if condition is None else jnp.hstack((y, condition))
        transformer_params = self.autoregressive_mlp(nn_input)
        transformer = self._flat_params_to_transformer(transformer_params)
        x = transformer.inverse(y)
        x = y.at[rank].set(x[rank])
        return (x, rank + 1), None

    def inverse_and_log_det(self, y, condition=None):
        x = self.inverse(y, condition)
        log_det = self.transform_and_log_det(x, condition)[1]
        return x, -log_det

    def _flat_params_to_transformer(self, params: Array):
        """Reshape to dim X params_per_dim, then vmap."""
        dim = self.shape[-1]  # type: ignore
        transformer_params = jnp.reshape(params, (dim, -1))
        transformer = eqx.filter_vmap(self.transformer_constructor)(transformer_params)
        return Batch(transformer, (dim,), vectorize_bijection=True)



"""Utility functions."""
from typing import Sequence

import equinox as eqx
import jax
import jax.numpy as jnp
from jax import Array
from jax.flatten_util import ravel_pytree
from jax.typing import ArrayLike


def real_to_increasing_on_interval(
    arr: Array, B: float = 1, softmax_adjust: float = 1e-2
):
    """Transform unconstrained parameter vector to monotonically increasing positions on
    [-B, B].

    Args:
        arr (Array): Parameter vector.
        B (float): Interval to transform output. Defaults to 1.
        softmax_adjust (float): Rescales softmax output using (widths +
            softmax_adjust/widths.size) / (1 + softmax_adjust). e.g. 0=no adjustment,
            1=average softmax output with evenly spaced widths, >1 promotes more evenly
            spaced widths.
    """
    if softmax_adjust < 0:
        raise ValueError("softmax_adjust should be >= 0.")
    widths = jax.nn.softmax(arr)  # type: ignore
    widths = (widths + softmax_adjust / widths.size) / (1 + softmax_adjust)
    widths = widths.at[0].set(widths[0] / 2)
    return 2 * B * jnp.cumsum(widths) - B


def inv_cum_sum(x):
    """Inverse of cumulative sum operation."""
    return x - jnp.pad(x[:-1], (1, 0))


def merge_cond_shapes(shapes: Sequence):
    """Merges shapes (tuples of ints or None) used in bijections and distributions.
    Returns None if all shapes are None, otherwise checks the shapes match, and returns
    the shape.
    """
    if len(shapes) == 0:
        raise ValueError("No shapes have been provided.")
    if all(s is None for s in shapes):
        return None
    shapes = [s for s in shapes if s is not None]
    if all(s == shapes[0] for s in shapes):
        return shapes[0]
    raise ValueError("The shapes do not match.")


def check_shapes_match(shapes: list[tuple[int, ...]]):
    """Check shapes match and produce a useful error message."""

    for i, shape in enumerate(shapes):
        if shape != shapes[0]:
            raise ValueError(
                f"Expected shapes to match, but index 0 had shape {shapes[0]}, and "
                f"index {i} had shape {shape}."
            )


def _get_ufunc_signature(in_shapes, out_shapes):
    """Convert a sequence of in_shapes and out_shapes to a universal function signature.

    Example:
        >>> _get_ufunc_signature([(3,),(2,3)], [()])
        "(3),(2,3)->()"
    """

    def _shapes_to_str(shapes):
        result = [str(s) if len(s) != 1 else str(s).replace(",", "") for s in shapes]
        return ",".join(result).replace(" ", "")

    in_shapes_str = _shapes_to_str(in_shapes)
    out_shapes_str = _shapes_to_str(out_shapes)
    return f"{in_shapes_str}->{out_shapes_str}"


def get_ravelled_bijection_constructor(bijection, filter_spec=eqx.is_inexact_array):
    """Given a bijection, returns a tuple containing
    1) a constructor for the bijection from a flattened array; 2) the current flattened
    parameters.

    Args:
        bijection (Bijection): Bijection.
        filter_spec: Filter function. Defaults to eqx.is_inexact_array.
    """
    params, static = eqx.partition(bijection, filter_spec)  # type: ignore
    current, unravel = ravel_pytree(params)

    def constructor(ravelled_params: Array):
        params = unravel(ravelled_params)
        return eqx.combine(params, static)

    return constructor, current


def arraylike_to_array(arr, err_name: str = "input", **kwargs) -> Array:
    """Combines jnp.asarray, with an isinstance(arr, ArrayLike) check. This
    allows inputs to be jax.numpy arrays, numpy arrays, python built in numeric types
    (float, int) etc, but does not allow list or tuple inputs (which are not arraylike
    and can introduce overhead and confusing behaviour in certain cases).

    Args:
        arr: Arraylike input to convert to a jax array.
        err_name (str, optional): Name of the input in the error message. Defaults to
            "input".
        **kwargs: Key word arguments passed to jnp.asarray.
    """
    if not isinstance(arr, ArrayLike):
        raise ValueError(
            f"Expected {err_name} to be arraylike; got {type(arr).__name__}."
        )
    return jnp.asarray(arr, **kwargs)



"""Utility bijections (embedding network, permutations, inversion etc.)"""
from typing import Callable

import jax.numpy as jnp
from jax import Array
from jax.experimental import checkify
from jax.typing import ArrayLike

from flowjax.bijections.bijection import Bijection


class Invert(Bijection):
    """Invert a bijection, such that the transform methods become the inverse
    methods and vice versa. Note that in general, we define bijections such that
    the forward methods are preffered, i.e. faster/actually implemented. For
    training flows, we generally want the inverse method (used in density
    evaluation), to be faster. Hence it is often useful to use this class to
    achieve this aim.
    """

    bijection: Bijection

    def __init__(self, bijection: Bijection):
        """
        Args:
            bijection (Bijection): Bijection to invert.
        """
        self.bijection = bijection
        self.shape = bijection.shape
        self.cond_shape = bijection.cond_shape

    def transform(self, x, condition=None):
        return self.bijection.inverse(x, condition)

    def transform_and_log_det(self, x, condition=None):
        return self.bijection.inverse_and_log_det(x, condition)

    def inverse(self, y, condition=None):
        return self.bijection.transform(y, condition)

    def inverse_and_log_det(self, y, condition=None):
        return self.bijection.transform_and_log_det(y, condition)


class Permute(Bijection):
    """Permutation transformation."""

    permutation: tuple[Array, ...]
    inverse_permutation: tuple[Array, ...]

    def __init__(self, permutation: ArrayLike):
        """
        Args:
            permutation (ArrayLike): An array with shape matching the array to
                transform, with elements 0-(array.size-1) representing the new order
                based on the flattened array (uses, C-like ordering).
        """
        permutation = jnp.asarray(permutation)
        checkify.check(
            (permutation.ravel().sort() == jnp.arange(permutation.size)).all(),
            "Invalid permutation array provided.",
        )
        self.shape = permutation.shape
        self.cond_shape = None

        indices = jnp.unravel_index(permutation.ravel(), permutation.shape)
        self.permutation = tuple(jnp.reshape(i, permutation.shape) for i in indices)

        inv_indices = jnp.unravel_index(
            jnp.argsort(permutation.ravel()), permutation.shape
        )
        self.inverse_permutation = tuple(
            jnp.reshape(i, permutation.shape) for i in inv_indices
        )

    def transform(self, x, condition=None):
        x, _ = self._argcheck_and_cast(x)
        return x[self.permutation]

    def transform_and_log_det(self, x, condition=None):
        x, _ = self._argcheck_and_cast(x)
        return x[self.permutation], jnp.array(0)

    def inverse(self, y, condition=None):
        y, _ = self._argcheck_and_cast(y)
        return y[self.inverse_permutation]

    def inverse_and_log_det(self, y, condition=None):
        x, _ = self._argcheck_and_cast(y)
        return y[self.inverse_permutation], jnp.array(0)


class Flip(Bijection):
    """Flip the input array. Condition argument is ignored."""

    def __init__(self, shape: tuple[int, ...]) -> None:
        """
        Args:
            shape (tuple[int, ...] | None): The shape of the bijection.
                Defaults to None.
        """
        self.shape = shape
        self.cond_shape = None

    def transform(self, x, condition=None):
        x, _ = self._argcheck_and_cast(x)
        return jnp.flip(x)

    def transform_and_log_det(self, x, condition=None):
        x, _ = self._argcheck_and_cast(x)
        return jnp.flip(x), jnp.array(0)

    def inverse(self, y, condition=None):
        y, _ = self._argcheck_and_cast(y)
        return jnp.flip(y)

    def inverse_and_log_det(self, y, condition=None):
        y, _ = self._argcheck_and_cast(y)
        return jnp.flip(y), jnp.array(0)


class Partial(Bijection):
    """Applies bijection to specific indices of an input."""

    bijection: Bijection
    idxs: int | slice | Array | tuple

    def __init__(self, bijection: Bijection, idxs, shape: tuple[int, ...]):
        """
        Args:
            bijection (Bijection): Bijection that is compatible with the subset of x
                indexed by idxs. idxs: Indices (Integer, a slice, or an ndarray with
                integer/bool dtype) of the transformed portion.
            shape (tuple[int, ...] | None): Shape of the bijection. Defaults to None.
        """
        self.bijection = bijection
        self.idxs = idxs
        self.shape = shape
        self.cond_shape = bijection.cond_shape

        if jnp.zeros(shape)[idxs].shape != bijection.shape:
            raise ValueError(
                f"The bijection shape is incompatible with the subset of the input "
                f"indexed by 'idxs'. The bijection has a shape of {bijection.shape}, "
                f"while the subset has a shape of {jnp.zeros(shape)[idxs].shape}."
            )

    def transform(self, x, condition=None):
        x, condition = self._argcheck_and_cast(x, condition)
        y = self.bijection.transform(x[self.idxs], condition)
        return x.at[self.idxs].set(y)

    def transform_and_log_det(self, x, condition=None):
        x, condition = self._argcheck_and_cast(x, condition)
        y, log_det = self.bijection.transform_and_log_det(x[self.idxs], condition)
        return x.at[self.idxs].set(y), log_det

    def inverse(self, y, condition=None):
        y, condition = self._argcheck_and_cast(y, condition)
        x = self.bijection.inverse(y[self.idxs], condition)
        return y.at[self.idxs].set(x)

    def inverse_and_log_det(self, y, condition=None):
        y, condition = self._argcheck_and_cast(y, condition)
        x, log_det = self.bijection.inverse_and_log_det(y[self.idxs], condition)
        return y.at[self.idxs].set(x), log_det


class EmbedCondition(Bijection):
    """Use an embedding network to reduce the dimensionality of the conditioning
    variable. The returned bijection has cond_dim equal to the raw condition size.
    """

    bijection: Bijection
    embedding_net: Callable

    def __init__(
        self,
        bijection: Bijection,
        embedding_net: Callable,
        raw_cond_shape: tuple[int, ...],
    ) -> None:
        """
        Args:
            bijection (Bijection): Bijection with ``bijection.cond_dim`` equal to the
                embedded size.
            embedding_net (Callable): A callable (e.g. equinox module) that embeds a
                conditioning variable to size ``bijection.cond_dim``.
            raw_cond_shape (tuple[int, ...] | None): The dimension of the raw
                conditioning variable.
        """
        self.bijection = bijection
        self.embedding_net = embedding_net

        self.shape = bijection.shape
        self.cond_shape = raw_cond_shape

    def transform(self, x, condition=None):
        x, condition = self._argcheck_and_cast(x, condition)
        condition = self.embedding_net(condition)
        return self.bijection.transform(x, condition)

    def transform_and_log_det(self, x, condition=None):
        x, condition = self._argcheck_and_cast(x, condition)
        condition = self.embedding_net(condition)
        return self.bijection.transform_and_log_det(x, condition)

    def inverse(self, y, condition=None):
        y, condition = self._argcheck_and_cast(y, condition)
        condition = self.embedding_net(condition)
        return self.bijection.inverse(y, condition)

    def inverse_and_log_det(self, y, condition=None):
        y, condition = self._argcheck_and_cast(y, condition)
        condition = self.embedding_net(condition)
        return self.bijection.inverse_and_log_det(y, condition)
